from .Animation import Animation
from .EasingCurve import EasingCurve
from .Chronoscope import Chronoscope
from .SequentialAnimation import SequentialAnimation
from .ParallelAnimation import ParallelAnimation



import glm
from glass import Instance


class AffineTransform(Instance):
    def __init__(self):
        Instance.__init__(self)

        self["affine_transform_row0"] = glm.vec4(1, 0, 0, 0)
        self["affine_transform_row1"] = glm.vec4(0, 1, 0, 0)
        self["affine_transform_row2"] = glm.vec4(0, 0, 1, 0)
        self["env_map_handle"] = 0
        self["visible"] = 1

    def apply(self, pos):
        transform_mat = glm.mat4(
            self["affine_transform_row0"],
            self["affine_transform_row1"],
            self["affine_transform_row2"],
            glm.vec4(0, 0, 0, 1),
        )
        transform_mat = glm.transpose(transform_mat)
        result = transform_mat * glm.vec4(*pos, 1)
        return result.xyz



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass.utils import checktype
from glass import FBO, ShaderProgram, sampler2D, GLConfig, GLInfo

from OpenGL import GL
import os
import time
import glm
from datetime import datetime


class ShaderEffect(PostProcessEffect):

    @checktype
    def __init__(
        self,
        shader_path: str = None,
        internal_format: GLInfo.internal_formats = GL.GL_RGBA32F,
        generate_mipmap: bool = False,
    ):
        PostProcessEffect.__init__(self)

        self._dynamic = False
        self._internal_format = internal_format
        self._generate_mipmap = generate_mipmap

        self._fbo = None
        self._program = None
        self._fragment_filename = ""
        self._uniforms = {}

        self._shader_path = ""
        self._start_time = 0
        self._last_frame_time = 0
        self._frame_index = 0

        if shader_path is not None:
            self.shader_path = shader_path

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D, self._internal_format)
        return self._fbo

    def __hash__(self):
        return id(self)

    @property
    def shader_path(self):
        return self._shader_path

    @shader_path.setter
    @checktype
    def shader_path(self, shader_path: str):
        shader_path = os.path.abspath(shader_path).replace("\\", "/")

        if not os.path.isfile(shader_path):
            raise FileNotFoundError(shader_path)

        if shader_path == self._shader_path:
            return

        self._shader_path = shader_path

    @property
    def program(self):
        if self._shader_path:
            if self._program is None:
                self_folder = os.path.dirname(os.path.abspath(__file__)).replace(
                    "\\", "/"
                )
                template_filename = (
                    self_folder
                    + "/../glsl/PostProcessEffects/shader_effect_template.fs"
                )
                self._program = ShaderProgram()
                self._program.define("FILE_NAME", f'"{self._shader_path}"')
                self._program.compile(Frame.draw_frame_vs)
                self._program.compile(template_filename)
            else:
                if (
                    "FILE_NAME" not in self._program.defines
                    or self._program.defines["FILE_NAME"] != f'"{self._shader_path}"'
                ):
                    self._program.define("FILE_NAME", f'"{self._shader_path}"')
                    self._program.reload()

        if self._uniforms and self._program is not None:
            for name, value in self._uniforms.items():
                self._program[name] = value
            self._uniforms.clear()

        return self._program

    def draw(self, screen_image: sampler2D) -> sampler2D:
        if self._generate_mipmap:
            screen_image.filter_mipmap = GL.GL_LINEAR

        if self.camera is not None:
            self.program["camera"] = self.camera

        self.program["depth_map"] = self.depth_map
        self.program["view_pos_map"] = self.view_pos_map
        self.program["view_normal_map"] = self.view_normal_map
        self.program["screen_image"] = screen_image
        self._dynamic = (
            self.program["iTime"].location != -1
            or self.program["iTimeDelta"].location != -1
            or self.program["iFrameRate"].location != -1
            or self.program["iFrame"].location != -1
            or self.program["iDate"].location != -1
        )

        if self._dynamic:
            current_time = time.time()
            now = datetime.now()

            if self._start_time == 0:
                self._start_time = current_time
            if self._last_frame_time == 0:
                self._last_frame_time = current_time

            time_delta = current_time - self._last_frame_time
            fps = 60
            if time_delta > 0:
                fps = 1 / time_delta
            t = current_time - self._start_time

            self.program["iTime"] = t
            self.program["iTimeDelta"] = time_delta
            self.program["iFrameRate"] = fps
            self.program["iFrame"] = self._frame_index
            self.program["iDate"] = glm.vec4(
                now.year, now.month, now.day, now.second + now.microsecond / 1000
            )

            self._last_frame_time = current_time
            self._frame_index += 1

        self.program.draw_triangles(start_index=0, total=6)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.draw(screen_image)

    def apply(self, screen_image: sampler2D) -> sampler2D:
        self.fbo.resize(screen_image.width, screen_image.height)
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            with self.fbo:
                self.draw(screen_image)

        return self.fbo.color_attachment(0)

    def __getitem__(self, name: str):
        if self._program is None:
            return self._uniforms[name]
        else:
            return self._program[name]

    def __setitem__(self, name: str, value):
        if self._program is None:
            self._uniforms[name] = value
        else:
            self.program[name] = value

    @property
    def should_update(self):
        if not self.enabled:
            return False

        return self._should_update or self._dynamic

    @should_update.setter
    def should_update(self, flag: bool):
        self._should_update = flag

    def need_pos_info(self) -> bool:
        return (
            self.program["view_pos_map"].location >= 0
            or self.program["view_normal_map"].location >= 0
            or self.program["depth_map"].location >= 0
        )



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex
from glass.ImageLoader import ImageLoader

import glm
import numpy as np
from enum import Enum


class ImageQuad(Mesh):

    class AlignCenter(Enum):
        Center = (0, 0)
        LeftCenter = (-0.5, 0)
        RightCenter = (0.5, 0)
        TopCenter = (0, 0.5)
        BottomCenter = (0, -0.5)
        LeftTop = (-0.5, 0.5)
        RightTop = (0.5, 0.5)
        LeftBottom = (-0.5, -0.5)
        RightBottom = (0.5, -0.5)

    @checktype
    def __init__(
        self,
        image: (str, np.ndarray),
        align_center: AlignCenter = AlignCenter.BottomCenter,
        width: float = None,
        height: float = None,
        vertical: bool = True,
        name: str = "",
    ):
        Mesh.__init__(
            self, color=glm.vec4(0), back_color=glm.vec4(0), name=name, block=True
        )
        self.__align_center = align_center

        _image = image
        aspect = 1
        if isinstance(image, str):
            try:
                _image = ImageLoader.load(image)
                aspect = _image.shape[1] / _image.shape[0]
            except:
                pass
        elif isinstance(image, np.ndarray):
            aspect = _image.shape[1] / _image.shape[0]

        if width is None and height is None:
            width = 1

        if width is None:
            width = height * aspect

        if height is None:
            height = width / aspect

        self.__width = width
        self.__height = height
        self.__vertical = vertical

        self.material.diffuse_map = image
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        width = self.__width
        height = self.__height
        vertical = self.__vertical
        align0 = self.__align_center.value[0] * width
        align1 = self.__align_center.value[1] * height

        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        # 左下
        vertex0 = Vertex()
        vertex0.position = glm.vec3(-width / 2 - align0, -height / 2 - align1, 0)
        if vertical:
            vertex0.position = glm.vec3(-width / 2 - align0, 0, -height / 2 - align1)
        vertex0.normal = normal
        vertex0.tex_coord = glm.vec3(0)

        # 右下
        vertex1 = Vertex()
        vertex1.position = glm.vec3(width / 2 - align0, -height / 2 - align1, 0)
        if vertical:
            vertex1.position = glm.vec3(width / 2 - align0, 0, -height / 2 - align1)
        vertex1.normal = normal
        vertex1.tex_coord = glm.vec3(1, 0, 0)

        # 右上
        vertex2 = Vertex()
        vertex2.position = glm.vec3(width / 2 - align0, height / 2 - align1, 0)
        if vertical:
            vertex2.position = glm.vec3(width / 2 - align0, 0, height / 2 - align1)
        vertex2.normal = normal
        vertex2.tex_coord = glm.vec3(1, 1, 0)

        # 左上
        vertex3 = Vertex()
        vertex3.position = glm.vec3(-width / 2 - align0, height / 2 - align1, 0)
        if vertical:
            vertex3.position = glm.vec3(-width / 2 - align0, 0, height / 2 - align1)
        vertex3.normal = normal
        vertex3.tex_coord = glm.vec3(0, 1, 0)

        vertices[0] = vertex0
        vertices[1] = vertex1
        vertices[2] = vertex2
        vertices[3] = vertex3

        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        self.generate_temp_TBN(vertices[0], vertices[1], vertices[2])
        self.generate_temp_TBN(vertices[0], vertices[2], vertices[3])

        del vertices[4:]
        del indices[2:]

    @property
    def vertical(self):
        return self.__vertical

    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag: bool):
        self.__vertical = flag

    @property
    def align_center(self):
        return self.__align_center

    @align_center.setter
    @Mesh.param_setter
    def align_center(self, align_center: AlignCenter):
        self.__align_center = align_center

    @property
    def width(self):
        return self.__width

    @width.setter
    @Mesh.param_setter
    def width(self, width: float):
        self.__width = width

    @property
    def height(self):
        return self.__height

    @height.setter
    @Mesh.param_setter
    def height(self, height: float):
        self.__height = height



from .Animation import Animation
from glass.DictList import DictList, ExtendableList

from typing import Any


def before_animation_add_callback(self: ExtendableList, animation: Any):
    if not isinstance(animation, Animation):
        raise TypeError("item add to AnimationGroup must be in type 'Animation'")


def after_animation_add_callback(self: ExtendableList, animation: Any):
    animation._parents.add(self)


def after_animation_remove_callback(self: ExtendableList, animation: Any):
    animation._parents.remove(self)


class AnimationGroup(Animation, DictList):

    def __init__(self, *animations, **kwargs):
        Animation.__init__(self, **kwargs)
        DictList.__init__(
            self,
            values=animations,
            before_item_add_callback=before_animation_add_callback,
            after_item_add_callback=after_animation_add_callback,
            after_item_remove_callback=after_animation_remove_callback,
        )

    def _update_duration(self):
        return

    @property
    def duration(self):
        self._update_duration()
        return self._duration

    @property
    def total_duration(self):
        self._update_duration()
        return self._total_duration



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math


def init_Tetrahedron(cls):
    sqrt_6 = math.sqrt(6)
    sqrt_2 = math.sqrt(2)
    sqrt_3 = math.sqrt(3)
    cls.edge_length = 2 / 3 * sqrt_6
    cls.base_positions = [
        glm.normalize(glm.vec3(2 / 3 * sqrt_2, 0, -1 / 3)),  # 0
        glm.normalize(glm.vec3(-sqrt_2 / 3, sqrt_6 / 3, -1 / 3)),  # 1
        glm.normalize(glm.vec3(-sqrt_2 / 3, -sqrt_6 / 3, -1 / 3)),  # 2
        glm.normalize(glm.vec3(0, 0, 1)),  # 3
    ]

    cls.base_indices = [
        glm.uvec3(0, 2, 1),
        glm.uvec3(3, 0, 1),
        glm.uvec3(3, 1, 2),
        glm.uvec3(3, 2, 0),
    ]

    cls.fixed_vertices = []
    for index in cls.base_indices:
        v1 = cls.base_positions[index[1]] - cls.base_positions[index[0]]
        v2 = cls.base_positions[index[2]] - cls.base_positions[index[0]]
        normal = glm.normalize(glm.cross(v1, v2))

        vertex1 = Vertex(
            position=cls.base_positions[index[0]],
            normal=normal,
            tex_coord=glm.vec3(0.5, 1, 0),
        )
        vertex2 = Vertex(
            position=cls.base_positions[index[1]],
            normal=normal,
            tex_coord=glm.vec3(0.5 - sqrt_3 / 4, 0.25, 0),
        )
        vertex3 = Vertex(
            position=cls.base_positions[index[2]],
            normal=normal,
            tex_coord=glm.vec3(0.5 + sqrt_3 / 4, 0.25, 0),
        )
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)
        cls.fixed_vertices.append(vertex3)

    return cls


@init_Tetrahedron
class Tetrahedron(Mesh):

    @checktype
    def __init__(
        self,
        radius=1,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        back_color: (glm.vec3, glm.vec4) = None,
        normalize_tex_coord=False,
        name: str = "",
    ):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        normalize_tex_coord = self.__normalize_tex_coord

        for i, fix_vert in enumerate(Tetrahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = 2 * radius * (
                    tex_coord - glm.vec3(0.5, 0.5, 0)
                ) * Tetrahedron.edge_length / math.sqrt(3) + glm.vec3(0.5, 0.5, 0)

            vertices[i] = Vertex(
                position=radius * fix_vert.position,
                normal=fix_vert.normal,
                tex_coord=tex_coord,
            )

        for i in range(0, len(Tetrahedron.fixed_vertices), 3):
            indices[int(i / 3)] = glm.uvec3(i, i + 1, i + 2)
            self.generate_temp_TBN(vertices[i], vertices[i + 1], vertices[i + 2])

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    @Mesh.param_setter
    def radius(self, radius: float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord

    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag: bool):
        self.__normalize_tex_coord = flag



from glass import ShaderProgram, sampler2D, GLConfig, sampler2DArray

import os
from OpenGL import GL


class Frame:

    def __init__(self):
        self._program = None

    @property
    def program(self):
        if self._program is None:
            self._program = ShaderProgram()
            self._program.compile(self.draw_frame_vs)
            self._program.compile(self.draw_frame_fs)
            self._program.uniform_not_set_warning = False

        return self._program

    @property
    def draw_frame_vs(self) -> str:
        self_folder = os.path.dirname(os.path.abspath(__file__))
        return os.path.abspath(self_folder + "/../glass/glsl/draw_frame.vs").replace(
            "\\", "/"
        )

    @property
    def draw_frame_fs(self) -> str:
        self_folder = os.path.dirname(os.path.abspath(__file__))
        return os.path.abspath(self_folder + "/glsl/Pipelines/draw_frame.fs").replace(
            "\\", "/"
        )

    def draw(
        self,
        screen_image: (sampler2D, sampler2DArray),
        gray: bool = False,
        invert: bool = False,
        layer: int = -1,
        index: int = 0,
    ):
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            if isinstance(screen_image, sampler2D):
                self.program["screen_image"] = screen_image
                self.program["layer"] = -1
            else:
                self.program["screen_image_array"] = screen_image
                self.program["layer"] = layer

            self.program["gray"] = gray
            self.program["invert"] = invert
            self.program["index"] = index
            self.program.draw_triangles(vertices=self.vertices, indices=self.indices)


Frame = Frame()



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm


class Box(Mesh):

    @checktype
    def __init__(
        self,
        Lx: float = 1,
        Ly: float = None,
        Lz: float = None,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        back_color: (glm.vec3, glm.vec4) = None,
        normalize_tex_coord=True,
        name: str = "",
    ):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        if Ly is None:
            Ly = Lx

        if Lz is None:
            Lz = Lx

        self.__Lx = Lx
        self.__Ly = Ly
        self.__Lz = Lz
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices

        Lx = self.__Lx
        Ly = self.__Ly
        Lz = self.__Lz
        normalize_tex_coord = self.__normalize_tex_coord

        # 左面
        position = glm.vec3(-Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.y, 0.5 + position.z, 0)
        )
        vertices[0] = Vertex(
            position=position, normal=glm.vec3(-1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.y, 0.5 + position.z, 0)
        )
        vertices[1] = Vertex(
            position=position, normal=glm.vec3(-1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.y, 0.5 + position.z, 0)
        )
        vertices[2] = Vertex(
            position=position, normal=glm.vec3(-1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.y, 0.5 + position.z, 0)
        )
        vertices[3] = Vertex(
            position=position, normal=glm.vec3(-1, 0, 0), tex_coord=tex_coord
        )

        # 右面
        position = glm.vec3(Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.y, 0.5 + position.z, 0)
        )
        vertices[4] = Vertex(
            position=position, normal=glm.vec3(1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.y, 0.5 + position.z, 0)
        )
        vertices[5] = Vertex(
            position=position, normal=glm.vec3(1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.y, 0.5 + position.z, 0)
        )
        vertices[6] = Vertex(
            position=position, normal=glm.vec3(1, 0, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.y, 0.5 + position.z, 0)
        )
        vertices[7] = Vertex(
            position=position, normal=glm.vec3(1, 0, 0), tex_coord=tex_coord
        )

        # 后面
        position = glm.vec3(-Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.z, 0)
        )
        vertices[8] = Vertex(
            position=position, normal=glm.vec3(0, -1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.z, 0)
        )
        vertices[9] = Vertex(
            position=position, normal=glm.vec3(0, -1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.z, 0)
        )
        vertices[10] = Vertex(
            position=position, normal=glm.vec3(0, -1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.z, 0)
        )
        vertices[11] = Vertex(
            position=position, normal=glm.vec3(0, -1, 0), tex_coord=tex_coord
        )

        # 前面
        position = glm.vec3(Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.x, 0.5 + position.z, 0)
        )
        vertices[12] = Vertex(
            position=position, normal=glm.vec3(0, 1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.x, 0.5 + position.z, 0)
        )
        vertices[13] = Vertex(
            position=position, normal=glm.vec3(0, 1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.x, 0.5 + position.z)
        )
        vertices[14] = Vertex(
            position=position, normal=glm.vec3(0, 1, 0), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 - position.x, 0.5 + position.z)
        )
        vertices[15] = Vertex(
            position=position, normal=glm.vec3(0, 1, 0), tex_coord=tex_coord
        )

        # 下面
        position = glm.vec3(-Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 - position.y)
        )
        vertices[16] = Vertex(
            position=position, normal=glm.vec3(0, 0, -1), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 - position.y)
        )
        vertices[17] = Vertex(
            position=position, normal=glm.vec3(0, 0, -1), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 - position.y)
        )
        vertices[18] = Vertex(
            position=position, normal=glm.vec3(0, 0, -1), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, -Ly / 2, -Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 - position.y)
        )
        vertices[19] = Vertex(
            position=position, normal=glm.vec3(0, 0, -1), tex_coord=tex_coord
        )

        # 上面
        position = glm.vec3(-Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.y)
        )
        vertices[20] = Vertex(
            position=position, normal=glm.vec3(0, 0, 1), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, -Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 0, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.y, 0)
        )
        vertices[21] = Vertex(
            position=position, normal=glm.vec3(0, 0, 1), tex_coord=tex_coord
        )

        position = glm.vec3(Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(1, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.y, 0)
        )
        vertices[22] = Vertex(
            position=position, normal=glm.vec3(0, 0, 1), tex_coord=tex_coord
        )

        position = glm.vec3(-Lx / 2, Ly / 2, Lz / 2)
        tex_coord = (
            glm.vec3(0, 1, 0)
            if normalize_tex_coord
            else glm.vec3(0.5 + position.x, 0.5 + position.y, 0)
        )
        vertices[23] = Vertex(
            position=position, normal=glm.vec3(0, 0, 1), tex_coord=tex_coord
        )

        # 左面
        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        # 右面
        indices[2] = 4 + glm.uvec3(0, 1, 2)
        indices[3] = 4 + glm.uvec3(0, 2, 3)

        # 后面
        indices[4] = 8 + glm.uvec3(0, 1, 2)
        indices[5] = 8 + glm.uvec3(0, 2, 3)

        # 前面
        indices[6] = 12 + glm.uvec3(0, 1, 2)
        indices[7] = 12 + glm.uvec3(0, 2, 3)

        # 下面
        indices[8] = 16 + glm.uvec3(0, 1, 2)
        indices[9] = 16 + glm.uvec3(0, 2, 3)

        # 上面
        indices[10] = 20 + glm.uvec3(0, 1, 2)
        indices[11] = 20 + glm.uvec3(0, 2, 3)

        del vertices[24:]
        del indices[12:]

        for index in indices:
            self.generate_temp_TBN(
                vertices[index[0]], vertices[index[1]], vertices[index[2]]
            )

    @property
    def Lx(self):
        return self.__Lx

    @Lx.setter
    @Mesh.param_setter
    def Lx(self, Lx: int):
        self.__Lx = Lx

    @property
    def Ly(self):
        return self.__Ly

    @Ly.setter
    @Mesh.param_setter
    def Ly(self, Ly: int):
        self.__Ly = Ly

    @property
    def Lz(self):
        return self.__Lz

    @Lz.setter
    @Mesh.param_setter
    def Lz(self, Lz: int):
        self.__Lz = Lz

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord

    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag: bool):
        self.__normalize_tex_coord = flag



from ..Mesh import Mesh
from ..ColorMap import ColorMap
from .Surf import Surf

from glass.utils import checktype

import numpy as np
import glm
import math


class CylindricalFSurf(Mesh):

    @checktype
    def __init__(
        self,
        func,
        r_range=[0, 3],
        theta_range=[0, 2 * math.pi],
        color_map: ColorMap = None,
        back_color_map: ColorMap = None,
        color: (glm.vec3, glm.vec4) = None,
        back_color: (glm.vec3, glm.vec4) = None,
        surf_type: Mesh.SurfType = Mesh.SurfType.Smooth,
        name: str = "",
    ):
        Mesh.__init__(self, name=name, surf_type=surf_type)
        self._func = func
        self._theta_range = theta_range
        self._r_range = r_range

        Surf._set_colors(self, color, back_color, color_map, back_color_map)
        self.start_building()

    def build(self):
        func = self._func
        theta_range = self._theta_range
        r_range = self._r_range

        theta = (
            np.linspace(theta_range[0], theta_range[1])
            if len(theta_range) == 2
            else theta_range
        )
        r = np.linspace(r_range[0], r_range[1]) if len(r_range) == 2 else r_range
        theta, r = np.meshgrid(theta, r)

        try:
            Z = func(r, theta)
            if Z.shape != theta.shape:
                raise Exception()
        except:
            func = np.vectorize(func)
            Z = func(r, theta)

        X = r * np.cos(theta)
        Y = r * np.sin(theta)
        C = Z

        Surf._build(self, X, Y, Z, C, C)

    @property
    def func(self):
        return self._func

    @func.setter
    @Mesh.param_setter
    def func(self, func):
        self._func = func

    @property
    def theta_range(self):
        return self._theta_range

    @theta_range.setter
    @Mesh.param_setter
    def theta_range(self, range):
        self._theta_range = range

    @property
    def r_range(self):
        return self._r_range

    @r_range.setter
    @Mesh.param_setter
    def r_range(self, range):
        self._r_range = range

    @property
    def color_map(self):
        return self._color_map

    @color_map.setter
    @Mesh.param_setter
    def color_map(self, color_map: ColorMap):
        self._color_map = color_map
        self._use_color_map = True
        if not self._back_color_map_user_set:
            self._back_color_map = color_map
            self._back_use_color_map = True

    @property
    def back_color_map(self):
        return self._back_color_map

    @back_color_map.setter
    @Mesh.param_setter
    def back_color_map(self, color_map: ColorMap):
        self._back_color_map = color_map
        self._back_use_color_map = True
        self._back_color_map_user_set = True



from ..Manipulators.Manipulator import Manipulator
from ..Renderers.Renderer import Renderer
from ..GlassEngineConfig import GlassEngineConfig
from ..PostProcessEffects import (
    PostProcessEffects,
    BloomEffect,
    ACESToneMapper,
    FXAAEffect,
    DOFEffect,
    SSAOEffect,
    ExplosureAdaptor,
)
from ..SlideAverageFilter import SlideAverageFilter
from ..VideoRecorder import VideoRecorder, convert_to_image
from ..Animations import Animation

from glass import (
    ShaderProgram,
    GLConfig,
    GlassConfig,
    FBO,
    RBO,
    sampler2DMS,
    sampler2D,
    RenderHints,
    SSBO,
    UBO,
    VAO,
)
from glass.utils import extname, di

import time
import glm
import sys
from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt
import os

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2
import numpy as np
import warnings


def __new__(cls, *args, **kwargs):
    if cls.qt.QtWidgets.QApplication.instance() is None:
        cls._app = cls.qt.QtWidgets.QApplication(sys.argv)

    instance = cls.__base__.__new__(cls, *args, **kwargs)
    return instance


def __init__(self, camera, parent=None) -> None:
    cls = self.__class__
    cls.__base__.__init__(self, parent=parent)

    self.setFocusPolicy(cls.qt.QtCore.Qt.FocusPolicy.StrongFocus)
    self._samples = 1
    self._samples_set_by_user = False

    self._pressed_keys = set()
    self._last_frame_time = 0
    self._is_cursor_hiden = False
    self._hide_cursor_global_pos = glm.vec2(0)
    self._last_cursor_global_pos = cls.qt.QtCore.QPoint(0, 0)
    self._fps = 0
    self._smooth_fps = 0
    self._before_draw_calls = 0
    self._before_draw_points = 0
    self._before_draw_lines = 0
    self._before_draw_meshes = 0
    self._before_draw_patches = 0
    self._draw_calls = 0
    self._draw_points = 0
    self._draw_lines = 0
    self._draw_meshes = 0
    self._draw_patches = 0
    self._paint_times = 0

    self._fps_filter = SlideAverageFilter()
    self._before_PPE_fbo_ss = None
    self._before_PPE_fbo_ms = None
    self._is_gl_init = False
    self._before_PPE_image = None
    self._video_recorders = []

    self._camera_id = id(camera)

    self._manipulator = None
    self._renderer = None

    self._render_hints = RenderHints()
    self._render_hints.depth_test = True

    self._listen_cursor_timer = self.startTimer(10)

    self._post_process_effects = PostProcessEffects()

    self._post_process_effects["SSAO"] = SSAOEffect()
    self._post_process_effects["bloom"] = BloomEffect()
    self._post_process_effects["DOF"] = DOFEffect()
    self._post_process_effects["explosure_adaptor"] = ExplosureAdaptor()
    self._post_process_effects["tone_mapper"] = ACESToneMapper()
    self._post_process_effects["FXAA"] = FXAAEffect(internal_format=GL.GL_RGBA8)

    self._post_process_effects["SSAO"].enabled = False
    self._post_process_effects["bloom"].enabled = False
    self._post_process_effects["DOF"].enabled = False
    self._post_process_effects["explosure_adaptor"].enabled = False
    self._post_process_effects["tone_mapper"].enabled = False
    self._post_process_effects["FXAA"].enabled = False


def update(self) -> None:
    self._before_PPE_image = None
    self.__class__.__base__.update(self)


def _assign_values_to_PPEs(self):
    self._post_process_effects.camera = self.camera
    self._post_process_effects.depth_map = self.renderer._depth_map
    self._post_process_effects.view_normal_map = self.renderer._view_normal_map
    self._post_process_effects.view_pos_map = self._renderer._view_pos_map


def _update(self) -> None:
    self.__class__.__base__.update(self)


def sizeHint(self):
    return self.__class__.qt.QtCore.QSize(800, 600)


@property
def camera(self):
    return di(self._camera_id)


@property
def samples(self) -> int:
    return self._samples


@samples.setter
def samples(self, samples: int) -> None:
    surface_format = self.__class__.qt.QtGui.QSurfaceFormat()
    surface_format.setSamples(samples)
    self.setFormat(surface_format)
    if not self._is_gl_init:
        self._samples = samples
        self._samples_set_by_user = True


def _set_samples(self, samples: int) -> None:
    surface_format = self.__class__.qt.QtGui.QSurfaceFormat()
    surface_format.setSamples(samples)
    self.setFormat(surface_format)
    if not self._is_gl_init:
        self._samples = samples


@property
def renderer(self) -> Renderer:
    return self._renderer


@renderer.setter
def renderer(self, renderer: Renderer) -> None:
    if self._renderer is renderer:
        return

    if renderer is None:
        self._renderer._camera_id = id(None)
        self._renderer = None
        return

    if renderer._camera_id != id(None):
        renderer.camera.screen.renderer = None

    self._renderer = renderer
    renderer._camera_id = self._camera_id
    should_update = renderer.startup()
    if should_update:
        self.update()


@property
def manipulator(self) -> Manipulator:
    return self._manipulator


@manipulator.setter
def manipulator(self, manipulator: Manipulator):
    if self._manipulator is manipulator:
        return

    if manipulator is None:
        self._manipulator._camera_id = id(None)
        self._manipulator = None
        return

    if manipulator._camera_id != id(None):
        manipulator.camera.screen.manipulator = None

    self._manipulator = manipulator
    manipulator._camera_id = self._camera_id
    should_update = manipulator.startup()
    if should_update:
        self.update()


@property
def render_hints(self) -> RenderHints:
    return self._render_hints


@render_hints.setter
def render_hints(self, hint: RenderHints) -> None:
    if hint is None:
        hint = RenderHints()

    self._render_hints = hint


def initializeGL(self) -> None:
    if GLConfig.major_version < 4:
        GlassEngineConfig["USE_SHADER_STORAGE_BLOCK"] = False

    if GLConfig.major_version == 4 and GLConfig.minor_version < 3:
        GlassEngineConfig["USE_SHADER_STORAGE_BLOCK"] = False

    if not bt.glInitBindlessTextureARB():
        if GlassConfig.warning:
            warning_message = """
Extension GL_ARB_bindless_texture is not available.
Shadows and dynamic environment mapping will be disabled.
Try to change default graphics card of python.exe to high performance one.
"""
            warnings.warn(warning_message)
        GlassEngineConfig["USE_BINDLESS_TEXTURE"] = False

    self.makeCurrent()

    if self.camera.scene is None:
        raise RuntimeError(
            f"{self.camera} is not in any scene, please add it to one scene before show it's screen"
        )

    self._is_gl_init = True


def makeCurrent(self) -> None:
    self.__class__.__base__.makeCurrent(self)
    GLConfig.buffered_current_context = GLConfig.current_context
    GLConfig.buffered_viewport = GLConfig.viewport

    SSBO.makeCurrent()
    UBO.makeCurrent()
    VAO.execute_cmd_buffer()


def resizeGL(self, width: int, height: int) -> None:
    self.__class__.__base__.resizeGL(self, width, height)
    self._before_PPE_image = None


def _draw_to_before_PPE(self, should_update_scene: bool) -> bool:
    if self._before_PPE_image is None or should_update_scene:
        with self._before_PPE_fbo:
            scene = self.camera.scene
            clear_color = scene.fog.apply(
                scene.background.color, glm.vec3(0), glm.vec3(0, self.camera.far, 0)
            )
            with GLConfig.LocalConfig(clear_color=clear_color):
                with self.render_hints:
                    should_update_scene = self.renderer.render() or should_update_scene

        resolved = self._before_PPE_fbo.resolved
        self._before_PPE_image = resolved.color_attachment(0)
        self._post_process_effects.screen_update_time = time.time()

    return should_update_scene


def _mark_draw_calls(self):
    self._before_draw_calls = ShaderProgram.accum_draw_calls()
    self._before_draw_points = ShaderProgram.accum_draw_points()
    self._before_draw_lines = ShaderProgram.accum_draw_lines()
    self._before_draw_meshes = ShaderProgram.accum_draw_meshes()
    self._before_draw_patches = ShaderProgram.accum_draw_patches()


def paintGL(self) -> None:
    self.makeCurrent()
    self.frame_started.emit()
    self._mark_draw_calls()

    camera = self.camera
    scene = camera.scene

    should_update_scene = scene.generate_meshes()
    should_update_PPEs = False

    if self._video_recorders:
        should_update_scene = self._draw_to_before_PPE(should_update_scene)
        if self._post_process_effects.has_valid:
            self._assign_values_to_PPEs()
            screen_image = self._post_process_effects.apply(self._before_PPE_image)
            should_update_PPEs = self._post_process_effects.should_update
        else:
            screen_image = self._before_PPE_image

        screen_image.fbo.draw_to_active(0)

        if self._paint_times > 0:
            view_port = GLConfig.buffered_viewport
            for video_recorder in self._video_recorders:
                video_recorder._start(view_port, self.smooth_fps)
                video_recorder._frame_queue.put((time.time(), screen_image.fbo.data(0)))
    elif self._post_process_effects.has_valid:
        should_update_scene = self._draw_to_before_PPE(should_update_scene)
        self._assign_values_to_PPEs()
        should_update_PPEs = self._post_process_effects.draw_to_active(
            self._before_PPE_image
        )
    else:
        clear_color = scene.fog.apply(
            scene.background.color, glm.vec3(0), glm.vec3(0, self.camera.far, 0)
        )
        with GLConfig.LocalConfig(clear_color=clear_color):
            with self.render_hints:
                should_update_scene = self.renderer.render() or should_update_scene

    self._calc_fps()
    self.frame_ended.emit()

    if not GlassEngineConfig.is_first_shown:
        GlassEngineConfig.is_first_shown = True
        for animation in Animation.all_instances:
            if animation._wait_to_start:
                animation._start()

    if should_update_scene or should_update_PPEs or Animation.has_valid():
        if should_update_scene:
            self._before_PPE_image = None

        self._update()


@property
def _before_PPE_fbo(self) -> FBO:
    screen_size = GLConfig.screen_size
    if self.samples > 1:
        if self._before_PPE_fbo_ms is not None:
            self._before_PPE_fbo_ms.resize(screen_size.x, screen_size.y, self.samples)
        else:
            self._before_PPE_fbo_ms = FBO(screen_size.x, screen_size.y, self.samples)
            self._before_PPE_fbo_ms.attach(0, sampler2DMS, GL.GL_RGBA32F)
            self._before_PPE_fbo_ms.attach(GL.GL_DEPTH_ATTACHMENT, RBO)
        return self._before_PPE_fbo_ms
    else:
        if self._before_PPE_fbo_ss is not None:
            self._before_PPE_fbo_ss.resize(screen_size.x, screen_size.y)
        else:
            self._before_PPE_fbo_ss = FBO(screen_size.x, screen_size.y)
            self._before_PPE_fbo_ss.attach(0, sampler2D, GL.GL_RGBA32F)
            self._before_PPE_fbo_ss.attach(GL.GL_DEPTH_ATTACHMENT, RBO)
        return self._before_PPE_fbo_ss


def _mouse_parameters(self, mouse_event) -> tuple:
    button = mouse_event.button()
    try:
        button_value = button.value
    except:
        button_value = int(button)

    button = Manipulator.MouseButton(button_value)
    try:
        screen_pos = mouse_event.position()
        global_pos = mouse_event.globalPosition()
    except:
        screen_pos = mouse_event.pos()
        global_pos = self.mapToGlobal(screen_pos)

    screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
    global_pos = glm.vec2(global_pos.x(), global_pos.y())
    return button, screen_pos, global_pos


def _wheel_parameters(self, wheel_event) -> tuple:
    angle = glm.vec2(wheel_event.angleDelta().x(), wheel_event.angleDelta().y())

    try:
        screen_pos = wheel_event.position()
        global_pos = wheel_event.globalPosition()
    except:
        screen_pos = wheel_event.pos()
        global_pos = self.mapToGlobal(screen_pos)

    screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
    global_pos = glm.vec2(global_pos.x(), global_pos.y())
    return angle, screen_pos, global_pos


def mousePressEvent(self, mouse_event) -> None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_pressed(
            button, screen_pos, global_pos
        )

    self.mouse_pressed.emit(button, screen_pos, global_pos)

    if should_update:
        self.update()


def mouseReleaseEvent(self, mouse_event) -> None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_released(
            button, screen_pos, global_pos
        )

    self.mouse_released.emit(button, screen_pos, global_pos)

    if should_update:
        self.update()


def mouseDoubleClickEvent(self, mouse_event) -> None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_double_clicked(
            button, screen_pos, global_pos
        )

    self.mouse_double_clicked.emit(button, screen_pos, global_pos)

    if should_update:
        self.update()


def _mouseMoveEvent(self, screen_pos: glm.vec2, global_pos: glm.vec2) -> bool:
    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_moved(screen_pos, global_pos)

    self.mouse_moved.emit(screen_pos, global_pos)

    if should_update:
        self.update()


def keyPressEvent(self, key_event) -> None:
    key = Manipulator.Key(key_event.key())
    if key not in self._pressed_keys:
        self._pressed_keys.add(key)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_pressed(key)

    self.key_pressed.emit(key)

    if should_update:
        self.update()


def keyReleaseEvent(self, key_event) -> None:
    key = Manipulator.Key(key_event.key())
    if key in self._pressed_keys:
        self._pressed_keys.remove(key)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_released(key)

    self.key_released.emit(key)

    if should_update:
        self.update()


def _keyRepeateEvent(self, keys: set) -> bool:
    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_repeated(keys)

    self.key_repeated.emit(keys)

    if should_update:
        self.update()


def wheelEvent(self, wheel_event) -> None:
    angle, screen_pos, global_pos = self._wheel_parameters(wheel_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_wheel_scrolled(
            angle, screen_pos, global_pos
        )

    self.wheel_scrolled.emit(angle, screen_pos, global_pos)

    if should_update:
        self.update()


def _calc_fps(self) -> None:
    current_time = time.time()
    dt = current_time - self._last_frame_time
    self._last_frame_time = current_time
    if dt < 1:
        self._fps = 1 / dt
        self._smooth_fps = self._fps_filter(self._fps)

    self._draw_calls = ShaderProgram.accum_draw_calls() - self._before_draw_calls
    self._draw_points = ShaderProgram.accum_draw_points() - self._before_draw_points
    self._draw_lines = ShaderProgram.accum_draw_lines() - self._before_draw_lines
    self._draw_meshes = ShaderProgram.accum_draw_meshes() - self._before_draw_meshes
    self._draw_patches = ShaderProgram.accum_draw_patches() - self._before_draw_patches
    self._paint_times += 1


def hide_cursor(self) -> None:
    if self._is_cursor_hiden:
        return

    self._is_cursor_hiden = True
    cursor_pos = self.__class__.qt.QtGui.QCursor.pos()
    self._hide_cursor_global_pos = glm.vec2(cursor_pos.x(), cursor_pos.y())
    self.setCursor(self.__class__.qt.QtCore.Qt.CursorShape.BlankCursor)
    return self._hide_cursor_global_pos


def show_cursor(self) -> None:
    if not self._is_cursor_hiden:
        return

    self._is_cursor_hiden = False
    self.unsetCursor()


def show(self) -> None:
    if self.__class__._app is not None and not self.__class__._has_exec:
        self.setWindowTitle("Glass Engine")
        self_folder = os.path.dirname(os.path.abspath(__file__))
        self.setWindowIcon(
            self.__class__.qt.QtGui.QIcon(
                self_folder + "/../images/glass_engine_logo64.png"
            )
        )

    self.__class__.__base__.show(self)
    if self.__class__._app is not None and not self.__class__._has_exec:
        self.__class__._app.exec()
        self.__class__._has_exec = True


@property
def post_process_effects(self) -> PostProcessEffects:
    return self._post_process_effects


@property
def is_cursor_hiden(self) -> bool:
    return self._is_cursor_hiden


@property
def fps(self) -> float:
    return self._fps


@property
def smooth_fps(self) -> float:
    return self._smooth_fps


@property
def draw_calls(self) -> int:
    return self._draw_calls


@property
def draw_points(self) -> int:
    return self._draw_points


@property
def draw_lines(self) -> int:
    return self._draw_lines


@property
def draw_meshes(self) -> int:
    return self._draw_meshes


@property
def draw_patches(self) -> int:
    return self._draw_patches


def timerEvent(self, timer_event) -> None:
    if timer_event.timerId() == self._listen_cursor_timer:
        cursor_global_pos = self.__class__.qt.QtGui.QCursor.pos()
        last_global_pos = self._last_cursor_global_pos
        self._last_cursor_global_pos = cursor_global_pos

        cursor_global_posF = self.__class__.qt.QtCore.QPointF(cursor_global_pos)

        should_update = False
        if self._pressed_keys:
            should_update = self._keyRepeateEvent(self._pressed_keys) or should_update

        if self._is_cursor_hiden:
            try:
                center_global_posF = self.mapToGlobal(
                    self.__class__.qt.QtCore.QPointF(self.width() / 2, self.height())
                )
            except:
                center = self.__class__.qt.QtCore.QPoint(
                    int(self.width() / 2), int(self.height() / 2)
                )
                center_global_pos = self.mapToGlobal(center)
                center_global_posF = self.__class__.qt.QtCore.QPointF(
                    float(center_global_pos.x()), float(center_global_pos.y())
                )

            offset = cursor_global_posF - center_global_posF
            buffer = 100
            if (
                offset.x() > buffer
                or offset.x() < -buffer
                or offset.y() > buffer
                or offset.y() < -buffer
            ):
                self._hide_cursor_global_pos.x -= offset.x()
                self._hide_cursor_global_pos.y -= offset.y()
                self.__class__.qt.QtGui.QCursor.setPos(center_global_posF.toPoint())
                cursor_global_posF = center_global_posF

        if cursor_global_pos != last_global_pos:
            try:
                screen_pos = self.mapFromGlobal(cursor_global_posF)
            except:
                screen_pos = self.mapFromGlobal(
                    self.__class__.qt.QtCore.QPoint(
                        int(cursor_global_posF.x()), int(cursor_global_posF.y())
                    )
                )

            screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
            global_pos = glm.vec2(cursor_global_posF.x(), cursor_global_posF.y())
            should_update = (
                self._mouseMoveEvent(screen_pos, global_pos) or should_update
            )

        if should_update:
            self.update()


def capture(self, save_path: str = None, viewport: tuple = None) -> np.ndarray:
    self.makeCurrent()
    with self._before_PPE_fbo:
        scene = self.camera.scene
        clear_color = scene.fog.apply(
            scene.background.color, glm.vec3(0), glm.vec3(0, self.camera.far, 0)
        )
        with GLConfig.LocalConfig(clear_color=clear_color):
            with self.render_hints:
                self.renderer.render()

    image = None
    resolved = self._before_PPE_fbo.resolved
    if self._post_process_effects.has_valid:
        self._before_PPE_image = resolved.color_attachment(0)
        self._post_process_effects.screen_update_time = time.time()
        self._assign_values_to_PPEs()
        result_sampler = self._post_process_effects.apply(self._before_PPE_image)
        image = result_sampler.fbo.data(0)
    else:
        image = resolved.data(0)

    if viewport is None:
        viewport = GLConfig.buffered_viewport
    image = convert_to_image(image, viewport)

    if save_path is not None:
        folder_name = os.path.dirname(os.path.abspath(save_path))
        if not os.path.isdir(folder_name):
            os.makedirs(folder_name)

        cv2.imwrite(save_path, image)

    return image


def capture_video(
    self, save_path: str, viewport: tuple = None, fps: (float, int) = None
) -> VideoRecorder:
    ext_name = extname(save_path)
    if ext_name not in ["mp4", "avi"]:
        raise ValueError(
            f"not supported video type: .{ext_name}, only support .mp4 and .avi"
        )

    fourcc = None
    if ext_name == "mp4":
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    elif ext_name == "avi":
        fourcc = cv2.VideoWriter_fourcc(*"XVID")

    folder_path = os.path.dirname(os.path.abspath(save_path))
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)

    if fps is None:
        fps = self.smooth_fps

    if viewport is None:
        self.__class__.__base__.makeCurrent(self)
        try:
            viewport = GLConfig.buffered_viewport
        except:
            viewport = None

    video_recorder = VideoRecorder(
        self._video_recorders, save_path, fourcc, viewport, fps
    )
    self._video_recorders.append(video_recorder)

    return video_recorder


def closeEvent(self, close_event) -> None:
    while self._video_recorders:
        video_recorder = self._video_recorders.pop()
        video_recorder.stop()


@property
def bloom(self):
    return self._post_process_effects["bloom"]


@bloom.setter
def bloom(self, flag: bool):
    self._post_process_effects["bloom"].enabled = flag


@property
def SSAO(self):
    return self._post_process_effects["SSAO"]


@SSAO.setter
def SSAO(self, flag: bool):
    self._post_process_effects["SSAO"].enabled = flag


@property
def tone_mapper(self):
    return self._post_process_effects["tone_mapper"]


@tone_mapper.setter
def tone_mapper(self, flag: bool):
    self._post_process_effects["tone_mapper"].enabled = flag


@property
def DOF(self):
    return self._post_process_effects["DOF"]


@DOF.setter
def DOF(self, flag: bool):
    self._post_process_effects["DOF"].enabled = flag


@property
def explosure_adaptor(self):
    return self._post_process_effects["explosure_adaptor"]


@explosure_adaptor.setter
def explosure_adaptor(self, flag: bool):
    self._post_process_effects["explosure_adaptor"].enabled = flag


@property
def FXAA(self):
    return self._post_process_effects["FXAA"]


@FXAA.setter
def FXAA(self, flag: bool):
    self._post_process_effects["FXAA"].enabled = flag


def init_QtScreen(cls):
    cls.qt = sys.modules[cls.__base__.__module__.split(".")[0]]
    cls._app = None
    cls._has_exec = False

    cls.__new__ = __new__
    cls.__init__ = __init__
    cls.update = update
    cls._assign_values_to_PPEs = _assign_values_to_PPEs
    cls._update = _update
    cls.sizeHint = sizeHint
    cls.camera = camera
    cls.samples = samples
    cls._set_samples = _set_samples
    cls.renderer = renderer
    cls.manipulator = manipulator
    cls.render_hints = render_hints
    cls.initializeGL = initializeGL
    cls.makeCurrent = makeCurrent
    cls.resizeGL = resizeGL
    cls._draw_to_before_PPE = _draw_to_before_PPE
    cls._mark_draw_calls = _mark_draw_calls
    cls.paintGL = paintGL
    cls._before_PPE_fbo = _before_PPE_fbo
    cls._mouse_parameters = _mouse_parameters
    cls._wheel_parameters = _wheel_parameters
    cls.mousePressEvent = mousePressEvent
    cls.mouseReleaseEvent = mouseReleaseEvent
    cls.mouseDoubleClickEvent = mouseDoubleClickEvent
    cls._mouseMoveEvent = _mouseMoveEvent
    cls.keyPressEvent = keyPressEvent
    cls.keyReleaseEvent = keyReleaseEvent
    cls._keyRepeateEvent = _keyRepeateEvent
    cls.wheelEvent = wheelEvent
    cls._calc_fps = _calc_fps
    cls.hide_cursor = hide_cursor
    cls.show_cursor = show_cursor
    cls.show = show
    cls.post_process_effects = post_process_effects
    cls.is_cursor_hiden = is_cursor_hiden
    cls.fps = fps
    cls.smooth_fps = smooth_fps
    cls.draw_calls = draw_calls
    cls.draw_points = draw_points
    cls.draw_lines = draw_lines
    cls.draw_meshes = draw_meshes
    cls.draw_patches = draw_patches
    cls.timerEvent = timerEvent
    cls.capture = capture
    cls.capture_video = capture_video
    cls.closeEvent = closeEvent
    cls.bloom = bloom
    cls.SSAO = SSAO
    cls.tone_mapper = tone_mapper
    cls.DOF = DOF
    cls.explosure_adaptor = explosure_adaptor
    cls.FXAA = FXAA

    return cls



from abc import ABC, abstractmethod
from glass import sampler2D
from glass.utils import checktype


class PostProcessEffect(ABC):

    def __init__(self):
        self._should_update = False
        self._enabled = True
        self._screen_update_time = 0

        self.depth_map = None
        self.view_pos_map = None
        self.view_normal_map = None
        self.camera = None

    def __bool__(self) -> bool:
        return self._enabled

    @abstractmethod
    def apply(self, screen_image: sampler2D) -> sampler2D:
        pass

    @abstractmethod
    def draw_to_active(self, screen_image: sampler2D) -> None:
        pass

    @abstractmethod
    def need_pos_info(self) -> bool:
        return False

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    @checktype
    def enabled(self, flag: bool):
        self._enabled = flag

    @property
    def should_update(self) -> bool:
        return self._should_update and self._enabled

    @should_update.setter
    @checktype
    def should_update(self, flag: bool):
        self._should_update = flag

    @property
    def screen_update_time(self):
        return self._screen_update_time

    @screen_update_time.setter
    @checktype
    def screen_update_time(self, screen_update_time: float):
        self._screen_update_time = screen_update_time



from glass.MetaInstancesRecorder import MetaInstancesRecorder

from .Chronoscope import Chronoscope
from .EasingCurve import EasingCurve
from glass.WeakSet import WeakSet
from ..GlassEngineConfig import GlassEngineConfig

import threading
import time


class Animation(metaclass=MetaInstancesRecorder):

    @MetaInstancesRecorder.init
    def __init__(
        self,
        target=None,
        property: str = "",
        from_=None,
        to=None,
        duration: float = 1,
        easing_curve: EasingCurve = EasingCurve.Linear,
        step: float = 0.01,
        loops: int = 1,
        go_back: bool = False,
        start_after_show: bool = True,
        running_callback=None,
        done_callback=None,
    ):
        self.target: object = target
        self.property: str = property
        self.from_: object = from_
        self.to: object = to
        self._duration: float = duration
        self._total_duration: float = 0
        self.__total_duration_dirty: bool = True
        self.step: float = step
        self._loops: int = loops
        self._go_back: bool = go_back
        self.start_after_show: bool = start_after_show
        self.easing_curve: EasingCurve = easing_curve
        self.running_callback = running_callback
        self.done_callback = done_callback

        self._timer = Chronoscope()
        self._running_thread = None
        self._wait_to_start: bool = False
        self._parents: WeakSet = WeakSet()

    @MetaInstancesRecorder.delete
    def __del__(self):
        pass

    def _start(self):
        if self._timer.status == Chronoscope.Status.Stop:
            self._running_thread = threading.Thread(
                target=self._run, args=(), daemon=True
            )
            self._running_thread.start()
        self._timer.start()
        self._wait_to_start: bool = False

    def start(self):
        if not GlassEngineConfig.is_first_shown and self.start_after_show:
            self._wait_to_start: bool = True
        else:
            self._start()

    def pause(self):
        self._timer.pause()
        self._wait_to_start: bool = False

    def stop(self, stop_value=None):
        self._running = False
        if self._running_thread is not None:
            if self._running_thread.is_alive():
                self._running_thread.join()

            self._running_thread = None

        self._timer.stop()
        if stop_value is None:
            stop_value = self.from_

        if self.target is not None and self.property:
            setattr(self.target, self.property, stop_value)

        self._wait_to_start: bool = False

    def _go_to(self, t: float):
        if t > self.total_duration:
            t = self.total_duration
            self._running = False

        progress = t / self.duration
        int_progress = int(progress)
        if progress != int_progress:
            progress -= int_progress
        elif progress != 0:
            progress = 1
        else:
            progress = 0

        if self.go_back and int_progress % 2 == 1:
            progress = 1 - progress

        mapped_progress = self.easing_curve(progress)

        if self.target is not None and self.property:
            if self.from_ is None:
                self.from_ = getattr(self.target, self.property)

            value = self.from_ + mapped_progress * (self.to - self.from_)
            setattr(self.target, self.property, value)

        if self.running_callback is not None:
            self.running_callback(t)

    def go_to(self, t: float):
        self._timer.go_to(t)
        self._go_to(t)

    @property
    def duration(self):
        return self._duration

    @property
    def _total_duration_dirty(self) -> bool:
        return self.__total_duration_dirty

    @_total_duration_dirty.setter
    def _total_duration_dirty(self, dirty: bool):
        if self.__total_duration_dirty == dirty:
            return

        self.__total_duration_dirty = dirty
        if dirty:
            for parent in self._parents:
                parent._total_duration_dirty = True

    @duration.setter
    def duration(self, duration: float):
        self._duration = duration
        self._total_duration_dirty = True

    @property
    def loops(self):
        return self._loops

    @loops.setter
    def loops(self, loops: int):
        self._loops = loops
        self._total_duration_dirty = True

    @property
    def go_back(self) -> bool:
        return self._go_back

    @go_back.setter
    def go_back(self, go_back: bool):
        self._go_back = go_back
        self._total_duration_dirty = True

    @property
    def speed(self):
        return self._timer.speed

    @speed.setter
    def speed(self, speed: float):
        self._timer.speed = speed

    @property
    def status(self):
        return self._timer.status

    @property
    def total_duration(self):
        if self._total_duration_dirty:
            self._total_duration = self.duration * self.loops
            if self.go_back:
                self._total_duration *= 2

            self._total_duration_dirty = False

        return self._total_duration

    @status.setter
    def status(self, status):
        if self._timer.status == status:
            return

        if self._timer.status == Chronoscope.Status.Stop:
            if status == Chronoscope.Status.Paused:
                self._timer.status = Chronoscope.Status.Paused
            else:
                self.start()
        elif self._timer.status == Chronoscope.Status.Paused:
            if status == Chronoscope.Status.Stop:
                self.stop()
            else:
                self.start()
        elif self._timer.status == Chronoscope.Status.Running:
            if status == Chronoscope.Status.Paused:
                self.pause()
            else:
                self.stop()

    def _run(self):
        self._running = True
        while self._running:
            current_time = time.perf_counter()
            next_time = current_time + self.step

            t = self._timer.time()
            self._go_to(t)

            if self._running:
                dt = next_time - time.perf_counter()
                if dt > 0:
                    time.sleep(dt)

        t = self._timer.time()
        if self.done_callback is not None:
            self.done_callback(t)
        self._timer.stop()

    @classmethod
    def has_valid(cls):
        for animation in cls.all_instances:
            if animation.status != Chronoscope.Status.Stop:
                return True

        return False



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass import FBO, ShaderProgram, sampler2D, GLConfig, Block
from glass.utils import checktype

from OpenGL import GL
import time
import os


class DOFEffect(PostProcessEffect):

    class CurrentFocus(Block.HostClass):
        def __init__(self):
            Block.HostClass.__init__(self)

            self._current_focus = 0

        @property
        def current_focus(self):
            return self._current_focus

        @current_focus.setter
        @Block.HostClass.not_const
        def current_focus(self, focus: float):
            self._current_focus = focus

    def __init__(self):
        PostProcessEffect.__init__(self)
        self.current_focus = DOFEffect.CurrentFocus()

        self._horizontal_fbo = None
        self._vertical_fbo = None
        self._program = None

    def need_pos_info(self) -> bool:
        return True

    @property
    def horizontal_fbo(self):
        if self._horizontal_fbo is None:
            self._horizontal_fbo = FBO()
            self._horizontal_fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        return self._horizontal_fbo

    @property
    def vertical_fbo(self):
        if self._vertical_fbo is None:
            self._vertical_fbo = FBO()
            self._vertical_fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        return self._vertical_fbo

    @property
    def program(self):
        if self._program is None:
            self._program = ShaderProgram()
            self._program.compile(Frame.draw_frame_vs)
            self._program.compile(
                os.path.dirname(os.path.abspath(__file__))
                + "/../glsl/PostProcessEffects/dof.fs"
            )
            self._program["CurrentFocus"].bind(self.current_focus)
        return self._program

    def apply(self, screen_image: sampler2D) -> sampler2D:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.program["camera"] = self.camera
            self.program["view_pos_map"] = self.view_pos_map
            self.program["fps"] = self.camera.screen.smooth_fps
            with self.horizontal_fbo:
                self.program["screen_image"] = screen_image
                self.program["horizontal"] = True
                self.program.draw_triangles(start_index=0, total=6)

            with self.vertical_fbo:
                self.program["screen_image"] = self.horizontal_fbo.color_attachment(0)
                self.program["horizontal"] = False
                self.program.draw_triangles(start_index=0, total=6)

        return self.vertical_fbo.color_attachment(0)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.program["camera"] = self.camera
            self.program["view_pos_map"] = self.view_pos_map
            self.program["fps"] = self.camera.screen.smooth_fps
            with self.horizontal_fbo:
                self.program["screen_image"] = screen_image
                self.program["horizontal"] = True
                self.program.draw_triangles(start_index=0, total=6)

            GLConfig.clear_buffers()
            self.program["screen_image"] = self.horizontal_fbo.color_attachment(0)
            self.program["horizontal"] = False
            self.program.draw_triangles(start_index=0, total=6)

    @property
    def should_update(self) -> bool:
        if not self._enabled:
            return False

        if not self.camera.lens.auto_focus:
            return False

        return (
            self.screen_update_time == 0
            or time.time() - self.screen_update_time
            <= self.camera.lens.focus_change_time + 1
        )

    @should_update.setter
    @checktype
    def should_update(self, flag: bool):
        pass



from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
from OpenGL import GL


class Points(Mesh):

    def __init__(
        self,
        points: list = [],
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        point_size: int = 5,
        name: str = "",
        block: bool = True,
    ):
        Mesh.__init__(
            self,
            primitive_type=GL.GL_POINTS,
            color=color,
            name=name,
            block=block,
            shared=False,
        )
        self.render_hints.point_size = point_size
        self.__points = points
        self.start_building()

    def build(self):
        points = self.__points
        vertices = self.vertices

        length = 0
        len_points = len(points)
        for i in range(len_points):
            if i > 0:
                length += glm.length(points[i] - points[i - 1])
            vertices[i] = Vertex(position=points[i], tex_coord=glm.vec3(length, 0, 0))

        del vertices[len_points:]

    @property
    def points(self):
        return self.__points

    @points.setter
    def points(self, points):
        self.__points = points
        self.start_building()

    @property
    def point_size(self):
        return self.render_hints.point_size

    @point_size.setter
    @checktype
    def point_size(self, point_size: int):
        self.render_hints.point_size = point_size



from .AnimationGroup import AnimationGroup


class ParallelAnimation(AnimationGroup):

    def __init__(self, *animations, **kwargs):
        AnimationGroup.__init__(self, *animations, **kwargs)

    def _update_duration(self):
        if not self._total_duration_dirty:
            return

        self._duration = 0
        for animation in self:
            if animation.total_duration > self._duration:
                self._duration = animation.total_duration

        self._total_duration = self._duration * self.loops
        if self._go_back:
            self._total_duration *= 2

        self._total_duration_dirty = False

    def _go_to(self, t: float):
        if t > self.total_duration:
            t = self.total_duration
            self._running = False

        progress = t / self.duration
        int_progress = int(progress)
        if progress != int_progress:
            progress -= int_progress
        elif progress != 0:
            progress = 1
        else:
            progress = 0

        if self.go_back and int_progress % 2 == 1:
            progress = 1 - progress

        reduce_t = progress * self.duration

        has_active_animation = False
        for animation in self:
            if animation is None:
                continue

            if reduce_t < animation.total_duration:
                animation._go_to(reduce_t)
                has_active_animation = True

        if not has_active_animation:
            self._running = False
            return

        if self.running_callback is not None:
            self.running_callback(t)



import os


def find_files(directory):
    file_list = []
    abs_directory = (
        os.path.abspath(os.path.dirname(os.path.abspath(__file__)) + "/..").replace(
            "\\", "/"
        )
        + "/"
        + directory
    )
    for root, _, files in os.walk(abs_directory):
        for file in files:
            file_path = os.path.abspath(os.path.join(root, file)).replace("\\", "/")
            if not os.path.isfile(file_path):
                continue

            target_path = (
                "glass_engine/"
                + directory
                + file_path[
                    len(abs_directory) : (-len(os.path.basename(file_path)) - 1)
                ]
            )
            if "__glcache__" not in file_path and "__pycache__" not in file_path:
                file_list.append((file_path.replace("\\", "/"), target_path))
    return file_list

self_folder = os.path.dirname(os.path.abspath(__file__))
datas = find_files("glsl")
datas.extend(
    [
        (
            os.path.abspath(self_folder + "/../images/glass_engine_logo64.png").replace("\\", "/"),
            "glass_engine/images",
        ),
        (
            os.path.abspath(self_folder + "/../LICENSE").replace("\\", "/"),
            "glass_engine",
        )
    ]
)

hiddenimports = [
    "glass_engine.Screens.PyQt5Screen",
    "glass_engine.Screens.PyQt6Screen",
    "glass_engine.Screens.PySide2Screen",
    "glass_engine.Screens.PySide6Screen",
    "glass_engine.Screens.QtScreen",
]



from .ShaderEffect import ShaderEffect
from glass import sampler2D, Block

import time
import os


class ExplosureAdaptor(ShaderEffect):

    class CurrentLuma(Block.HostClass):
        def __init__(self):
            Block.HostClass.__init__(self)
            self._current_luma = 0

        @property
        def current_luma(self):
            return self._current_luma

        @current_luma.setter
        @Block.HostClass.not_const
        def current_luma(self, luma: float):
            self._current_luma = luma

    def __init__(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(
            self,
            self_folder + "/../glsl/PostProcessEffects/explosure_adaptor.glsl",
            generate_mipmap=True,
        )

        self.current_luma = ExplosureAdaptor.CurrentLuma()
        self["CurrentLuma"] = self.current_luma

    def apply(self, screen_image: sampler2D) -> sampler2D:
        self.program["fps"] = self.camera.screen.smooth_fps
        return ShaderEffect.apply(self, screen_image)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        self.program["fps"] = self.camera.screen.smooth_fps
        ShaderEffect.draw_to_active(self, screen_image)

    @property
    def should_update(self) -> bool:
        if not self._enabled:
            return False

        if (not self.camera.lens.auto_explosure) or self.camera.lens.local_explosure:
            return False

        return (
            self.screen_update_time == 0
            or time.time() - self.screen_update_time
            <= self.camera.lens.explosure_adapt_time
        )

    @should_update.setter
    def should_update(self, flag: bool):
        pass



from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
import math
from OpenGL import GL


class Circle(Mesh):

    def __init__(
        self,
        radius: float = 1,
        vertical: bool = False,
        n_points: int = 100,
        start_angle: float = 0,
        span_angle: float = 360,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        line_width: int = 2,
        name: str = "",
    ):
        Mesh.__init__(self, primitive_type=GL.GL_LINE_STRIP, color=color, name=name)
        self.render_hints.line_width = line_width
        self.__radius: float = radius
        self.__vertical: bool = vertical
        self.__n_points: int = n_points
        self.__start_angle: float = start_angle
        self.__span_angle: float = span_angle
        self.start_building()

    def build(self):
        vertices = self.vertices
        radius: float = self.__radius
        vertical: bool = self.__vertical
        n_points: int = self.__n_points
        start_angle: float = self.__start_angle
        span_angle: float = self.__span_angle

        for i in range(n_points):
            theta = (start_angle + i / (n_points - 1) * span_angle) / 180 * math.pi

            position = radius * glm.vec3(math.cos(theta), math.sin(theta), 0)
            bitangent = glm.vec3(-math.sin(theta), math.cos(theta), 0)
            if vertical:
                position = radius * glm.vec3(math.cos(theta), 0, math.sin(theta))
                bitangent = glm.vec3(-math.sin(theta), 0, math.cos(theta))

            vertices[i] = Vertex(
                position=position,
                bitangent=bitangent,
                tex_coord=glm.vec3(radius * theta, 0, 0),
            )

        del vertices[n_points:]

    @property
    def line_width(self) -> int:
        return self.render_hints.line_width

    @line_width.setter
    @checktype
    def line_width(self, line_width: int) -> None:
        self.render_hints.line_width = line_width

    @property
    def radius(self) -> float:
        return self.__radius

    @radius.setter
    @Mesh.param_setter
    def radius(self, radius: float) -> None:
        self.__radius = radius

    @property
    def vertical(self) -> bool:
        return self.__vertical

    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag: bool) -> None:
        self.__vertical = flag

    @property
    def n_points(self) -> int:
        return self.__n_points

    @n_points.setter
    @Mesh.param_setter
    def n_points(self, n_points: int) -> None:
        self.__n_points = n_points

    @property
    def start_angle(self) -> float:
        return self.__start_angle

    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, angle: float) -> None:
        self.__start_angle = angle

    @property
    def span_angle(self) -> float:
        return self.__span_angle

    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, angle: float) -> None:
        self.__span_angle = angle



from .Light import Light, FlatLight
from glass.DictList import DictList
from glass import Block

import glm


class DirLight(Light):
    pass


class FlatDirLight(FlatLight):

    def __init__(self, dir_light: DirLight):
        self.direction = glm.vec3(0, 1, 0)
        self.abs_orientation = glm.quat(1, 0, 0, 0)
        self.max_back_offset = 0
        FlatLight.__init__(self, dir_light)


class DirLights(Block.HostClass):

    def __init__(self):
        Block.HostClass.__init__(self)
        self.dir_lights = DictList()

    def __getitem__(self, key: (str, int)):
        return self.dir_lights[key]

    @Block.HostClass.not_const
    def __setitem__(self, key: (str, int), dir_light: FlatDirLight):
        self.dir_lights[key] = dir_light

    @Block.HostClass.not_const
    def __delitem__(self, key: (str, int)):
        del self.dir_lights[key]

    def __contains__(self, path_str: str):
        return path_str in self.dir_lights

    def __len__(self):
        return len(self.dir_lights)

    def __iter__(self):
        return iter(self.dir_lights)

    def keys(self):
        return self.dir_lights.keys()

    @property
    def n_dir_lights(self):
        return len(self.dir_lights)



from .Manipulator import Manipulator

from glass.RenderHints import RenderHints
from glass.utils import checktype

from OpenGL import GL
import math
import glm
import datetime


class ModelViewManipulator(Manipulator):

    def __init__(self, distance: float = 5, azimuth: float = 0, elevation: float = 0):
        Manipulator.__init__(self)

        self.__is_left_pressed = False
        self.__left_press_global_posF = glm.vec2(0, 0)
        self.__left_press_azimuth = 0
        self.__left_press_elevation = 0

        self.__is_right_pressed = False
        self.__right_press_global_posF = glm.vec2(0, 0)
        self.__right_press_offset = glm.vec2(0, 0)

        self.__distance = distance
        self.__azimuth = azimuth
        self.__elevation = elevation
        self.__offset = glm.vec2(0, 0)
        self.drag_sensitivity: float = 1
        self.scroll_sensitivity: float = 1

    @property
    def distance(self) -> float:
        return self.__distance

    @distance.setter
    @checktype
    def distance(self, distance: float):
        self.__distance = distance
        self.__update_camera()
        self.camera.screen.update()

    @property
    def azimuth(self):
        return self.__azimuth

    @azimuth.setter
    @checktype
    def azimuth(self, azimuth: float):
        self.__azimuth = azimuth
        self.__update_camera()
        self.camera.screen.update()

    @property
    def elevation(self):
        return self.__elevation

    @elevation.setter
    @checktype
    def elevation(self, elevation: float):
        self.__elevation = elevation
        self.__update_camera()
        self.camera.screen.update()

    def __update_camera(self):
        azimuth = self.__azimuth / 180 * math.pi
        elevation = self.__elevation / 180 * math.pi

        self.camera.position.x = (
            self.__distance * math.cos(elevation) * math.sin(azimuth)
        )
        self.camera.position.y = (
            -self.__distance * math.cos(elevation) * math.cos(azimuth)
        )
        self.camera.position.z = self.__distance * math.sin(elevation)

        if glm.length(self.__offset) > 1e-6:
            right = glm.vec3(math.cos(azimuth), math.sin(azimuth), 0)
            forward = glm.vec3(
                -math.sin(elevation) * math.sin(azimuth),
                math.sin(elevation) * math.cos(azimuth),
                math.cos(elevation),
            )
            self.camera.position += self.__offset.x * right + self.__offset.y * forward

        self.camera.pitch = -self.__elevation
        self.camera.yaw = self.__azimuth

    def startup(self):
        self.__update_camera()

    def on_mouse_pressed(
        self,
        button: Manipulator.MouseButton,
        screen_pos: glm.vec2,
        global_pos: glm.vec2,
    ):
        if button == Manipulator.MouseButton.LeftButton:
            self.__is_left_pressed = True
            self.__left_press_global_posF = global_pos
            self.__left_press_azimuth = self.__azimuth
            self.__left_press_elevation = self.__elevation
        elif button == Manipulator.MouseButton.RightButton:
            self.__is_right_pressed = True
            self.__right_press_global_posF = global_pos
            self.__right_press_offset = self.__offset
        elif button == Manipulator.MouseButton.XButton1:
            self.drag_sensitivity /= pow(2, 1 / 2)
        elif button == Manipulator.MouseButton.XButton2:
            self.drag_sensitivity *= pow(2, 1 / 2)

        return False

    def on_mouse_released(
        self,
        button: Manipulator.MouseButton,
        screen_pos: glm.vec2,
        global_pos: glm.vec2,
    ):
        if button == Manipulator.MouseButton.LeftButton:
            self.__is_left_pressed = False
            if self.__left_press_global_posF == global_pos:
                x = screen_pos.x
                y = screen_pos.y
                width = self.camera.screen.width()
                height = self.camera.screen.height()
                s = x / (width - 1)
                t = 1 - y / (height - 1)
                self.camera.lens.focus_tex_coord = glm.vec2(s, t)
                return True
        elif button == Manipulator.MouseButton.RightButton:
            self.__is_right_pressed = False

        return False

    def on_mouse_double_clicked(
        self,
        button: Manipulator.MouseButton,
        screen_pos: glm.vec2,
        global_pos: glm.vec2,
    ) -> bool:
        if button == Manipulator.MouseButton.LeftButton:
            self.__offset = glm.vec2(0, 0)
            self.camera.fov = 45
            self.__update_camera()
            return True

        return False

    def on_mouse_moved(self, screen_pos: glm.vec2, global_pos: glm.vec2):
        if self.__is_left_pressed:
            d = global_pos - self.__left_press_global_posF
            dx = d.x
            dy = d.y

            d_pitch = dy / self.camera.screen.height() * 200 * self.drag_sensitivity
            d_yaw = -dx / self.camera.screen.width() * 200 * self.drag_sensitivity

            self.__azimuth = self.__left_press_azimuth + d_yaw
            self.__elevation = self.__left_press_elevation + d_pitch
            self.__update_camera()

            return True

        elif self.__is_right_pressed:
            d = global_pos - self.__right_press_global_posF
            dx = -d.x
            dy = d.y

            offset_delta = glm.vec2()
            offset_delta.x = (
                dx
                / self.camera.screen.width()
                * 2
                * self.__distance
                * self.camera.tan_half_fov
                * self.camera.aspect
            )
            offset_delta.y = (
                dy
                / self.camera.screen.height()
                * 2
                * self.__distance
                * self.camera.tan_half_fov
            )
            self.__offset = self.__right_press_offset + offset_delta
            self.__update_camera()

            return True

        return False

    def on_wheel_scrolled(
        self, angle: glm.vec2, screen_pos: glm.vec2, global_pos: glm.vec2
    ):
        n = self.scroll_sensitivity * angle.y / 120
        scale = pow(2, n / 6)
        if self.camera.projection_mode.value == 0:
            self.camera.fov /= scale
        else:
            self.camera.height /= scale

        return True

    def on_key_pressed(self, key: Manipulator.Key) -> bool:
        if key == Manipulator.Key.Key_R:
            polygon_mode = self.camera.screen.render_hints.polygon_mode
            if polygon_mode in [GL.GL_FILL, RenderHints.inherit]:
                self.camera.screen.render_hints.polygon_mode = GL.GL_LINE
                self.camera.screen.render_hints.line_width = 1
            elif polygon_mode == GL.GL_LINE:
                self.camera.screen.render_hints.polygon_mode = GL.GL_POINT
                self.camera.screen.render_hints.point_size = 1.5
            elif polygon_mode == GL.GL_POINT:
                self.camera.screen.render_hints.polygon_mode = GL.GL_FILL
            return True
        elif key == Manipulator.Key.Key_P:
            now = datetime.datetime.now()
            file_name = "capture_" + now.strftime("%Y-%m-%d_%H-%M-%S") + ".png"
            self.camera.screen.capture(file_name)
        elif key == Manipulator.Key.Key_O:
            self.camera.screen.SSAO.enabled = not self.camera.screen.SSAO.enabled
            return True
        elif key == Manipulator.Key.Key_M:
            self.camera.screen.DOF.enabled = not self.camera.screen.DOF.enabled
            return True

    def on_key_repeated(self, keys) -> bool:
        should_update = False
        if Manipulator.Key.Key_W in keys and Manipulator.Key.Key_S not in keys:
            self.__distance /= pow(2, 1 / self.camera.screen.smooth_fps)
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_S in keys and Manipulator.Key.Key_W not in keys:
            self.__distance *= pow(2, 1 / self.camera.screen.smooth_fps)
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_A in keys and Manipulator.Key.Key_D not in keys:
            self.__offset.x -= (
                0.5
                / self.camera.screen.smooth_fps
                * self.__distance
                * self.camera.tan_half_fov
            )
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_D in keys and Manipulator.Key.Key_A not in keys:
            self.__offset.x += (
                0.5
                / self.camera.screen.smooth_fps
                * self.__distance
                * self.camera.tan_half_fov
            )
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_E in keys and Manipulator.Key.Key_C not in keys:
            self.__offset.y += (
                0.5
                / self.camera.screen.smooth_fps
                * self.__distance
                * self.camera.tan_half_fov
            )
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_C in keys and Manipulator.Key.Key_E not in keys:
            self.__offset.y -= (
                0.5
                / self.camera.screen.smooth_fps
                * self.__distance
                * self.camera.tan_half_fov
            )
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_F in keys:
            screen = self.camera.screen
            print("fps:", screen.fps, "draw calls:", screen.draw_calls)
            should_update = True

        return should_update



import glm
import math
import uuid
import numpy as np

from glass.DictList import DictList
from glass.WeakSet import WeakSet
from glass.WeakDict import WeakDict
from glass.MetaInstancesRecorder import MetaInstancesRecorder
from .callback_vec import callback_quat, callback_vec3


class SceneNode(metaclass=MetaInstancesRecorder):

    @MetaInstancesRecorder.init
    def __init__(self, name: str = "", unique_path: bool = False):
        if name:
            self._name = name
        else:
            self._name = self.__class__.__name__ + "_" + str(uuid.uuid1())

        self._unique_path: bool = unique_path
        self._position: callback_vec3 = callback_vec3(0, 0, 0, callback=self._set_dirty)
        self._orientation: callback_quat = callback_quat(
            1, 0, 0, 0, callback=self._update_yaw_pitch_roll
        )
        self._scale: callback_vec3 = callback_vec3(1, 1, 1, callback=self._set_dirty)
        self._yaw_pitch_roll: glm.vec3 = glm.vec3(0, 0, 0)

        self.abs_position = glm.vec3(0, 0, 0)
        self.abs_orientation = glm.quat(1, 0, 0, 0)

        self._parents: DictList = DictList(weak_ref=True)
        self._children: DictList = DictList()
        self._scenes: WeakSet = WeakSet()
        self._transform_dirty: WeakSet = WeakSet()
        self._children_transform_dirty: WeakDict = WeakDict(
            weak_ref_keys=True, weak_ref_values=False
        )
        self._should_update_yaw_pitch_roll: bool = True

        self._propagation_props: dict = {}
        self._block_propagation: set = set()

        self._propagation_props["visible"] = True

    @MetaInstancesRecorder.delete
    def __del__(self):
        pass

    @property
    def unique_path(self) -> bool:
        return self._unique_path

    @unique_path.setter
    def unique_path(self, flag: bool) -> None:
        self._unique_path = flag

    @property
    def visible(self):
        return self.propagation_prop("visible")

    @visible.setter
    def visible(self, visible: bool):
        self.set_propagation_prop("visible", visible)

    def hide(self):
        self.visible = False

    def show(self):
        self.visible = True

    @property
    def scenes(self):
        return self._scenes

    @property
    def parents(self):
        return self._parents

    @property
    def children(self):
        return self._children

    def block_propagation(self, name: str, flag: bool = True):
        if flag:
            self._block_propagation.add(name)
        elif name in self._block_propagation:
            self._block_propagation.remove(name)

    def propagation_prop(self, name: str):
        if name not in self._propagation_props:
            return None

        return self._propagation_props[name]

    def set_propagation_prop(
        self, name: str, value, callback=None, args: tuple = (), kwargs: dict = {}
    ):
        parent_queues = [self]
        is_self_set = False
        while parent_queues:
            parent = parent_queues.pop()
            if name in parent._block_propagation:
                child_queues = [parent]
                while child_queues:
                    child = child_queues.pop()
                    child._propagation_props[name] = value
                    if callback is not None:
                        callback(child, *args, **kwargs)
                    child_queues.extend(child._children)
                    if child is self:
                        is_self_set = True
            else:
                parent_queues.extend(parent.parents)

        if not is_self_set:
            child_queues = [self]
            while child_queues:
                child = child_queues.pop()
                child._propagation_props[name] = value
                if callback is not None:
                    callback(child, *args, **kwargs)
                child_queues.extend(child._children)

    def has_parent(self, parent):
        return parent in self._parents

    def has_child(self, child):
        return child in self._children

    def add_child(self, node):
        if node.has_parent(self):
            return

        if node.name in self._children:
            if self._children[node.name] is not node:
                old_name = node.name
                i = 1
                new_name = f"{old_name}-{i}"
                while new_name in self._children:
                    i += 1
                    new_name = f"{old_name}-{i}"
                node.name = new_name
            else:
                return

        if self.name in node._parents:
            if node._parents[self.name] is not self:
                raise NameError("already has one parent named '" + self.name + "'")
            else:
                return

        node._parents[self.name] = self
        self._children[node.name] = node
        node._add_as_child_callback()

        node._add_scenes(self._scenes)
        self._set_dirty(False, True)

    def remove_child(self, child):
        detached_node = None
        paths_prefix = None
        if isinstance(child, str):
            name = child
            if name in self._children:
                detached_node = self._children[name]
                paths_prefix = detached_node.paths_str
                del self._children[name]
                del detached_node._parents[self.name]
        else:
            try:
                self._children.remove(child)
                paths_prefix = child.paths_str
                child._parents.remove(self)
                detached_node = child
            except ValueError:
                pass

        if detached_node is None:
            return None

        detached_node._update_scenes()
        for scene in self.scenes:
            scene._remove_paths_prefix(paths_prefix)

    def clear_children(self):
        children_names = self.children_names
        for child_name in children_names:
            self.remove_child(child_name)

    @property
    def children_names(self):
        return list(self._children.keys())

    def __hash__(self):
        return id(self)

    def __getitem__(self, name: (str, int)):
        if isinstance(name, str):
            if name not in self._children:
                self.add_child(SceneNode(name))
        else:
            while name >= len(self._children):
                self.add_child(SceneNode())

        return self._children[name]

    def __setitem__(self, name: str, node):
        if name in self._children:
            if self._children[name] is node:
                return
            self.remove_child(name)

        node.name = name
        self.add_child(node)

    def __delitem__(self, name: str):
        self.remove_child(name)

    def __to_string(self, indent):
        result = indent * "  " + self.name
        for child in self._children:
            result += "\n" + child.__to_string(indent + 1)
        return result

    def __repr__(self):
        return self.__to_string(0)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        if name == self._name:
            return

        if not self._parents:
            self._name = name
            return

        for parent in self._parents:
            if name in parent._children:
                raise NameError(
                    "parent node already has one child named '" + name + "'"
                )

            del parent._children[self._name]
            parent._children[name] = self

        self._name = name

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position: glm.vec3):
        self._position.x = position.x
        self._position.y = position.y
        self._position.z = position.z

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale: (glm.vec3, float)):
        if isinstance(scale, (int, float)):
            self._scale.x = scale
            self._scale.y = scale
            self._scale.z = scale
        else:
            self._scale.x = scale.x
            self._scale.y = scale.y
            self._scale.z = scale.z

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation: glm.quat):
        self._orientation.w = orientation.w
        self._orientation.x = orientation.x
        self._orientation.y = orientation.y
        self._orientation.z = orientation.z

    def rotate(self, axis:glm.vec3, angle:float):
        angle_rad = angle/180*math.pi
        new_orientation = glm.quat(math.cos(angle_rad/2), math.sin(angle_rad/2)*axis) * self._orientation
        self._orientation.w = new_orientation.w
        self._orientation.x = new_orientation.x
        self._orientation.y = new_orientation.y
        self._orientation.z = new_orientation.z

    def translate(self, translation:glm.vec3):
        self._position.x = self._position.x + translation.x
        self._position.y = self._position.x + translation.y
        self._position.z = self._position.x + translation.z

    def _add_scenes(self, scenes):
        self._scenes.update(scenes)
        for child in self._children:
            child._add_scenes(scenes)

    def _update_yaw_pitch_roll(self):
        if not self._should_update_yaw_pitch_roll:
            return

        q0 = self._orientation.w
        q1 = self._orientation.x
        q2 = self._orientation.y
        q3 = self._orientation.z
        self._yaw_pitch_roll[0] = (
            math.atan2(2 * (q0 * q3 - q1 * q2), 1 - 2 * (q1 * q1 + q3 * q3))
            / math.pi
            * 180
        )
        self._yaw_pitch_roll[1] = (
            math.asin(np.clip(2 * (q0 * q1 + q2 * q3), -1, 1)) / math.pi * 180
        )
        self._yaw_pitch_roll[2] = (
            math.atan2(2 * (q0 * q2 - q1 * q3), 1 - 2 * (q1 * q1 + q2 * q2))
            / math.pi
            * 180
        )

        self._set_dirty()

    def _update_orientation(self):
        yaw = self._yaw_pitch_roll[0] / 180 * math.pi
        pitch = self._yaw_pitch_roll[1] / 180 * math.pi
        roll = self._yaw_pitch_roll[2] / 180 * math.pi

        quat1 = glm.quat(math.cos(yaw / 2), 0, 0, math.sin(yaw / 2))
        quat2 = glm.quat(math.cos(pitch / 2), math.sin(pitch / 2), 0, 0)
        quat3 = glm.quat(math.cos(roll / 2), 0, math.sin(roll / 2), 0)

        self._should_update_yaw_pitch_roll = False
        self.orientation = quat1 * quat2 * quat3
        self._should_update_yaw_pitch_roll = True

        self._set_dirty()

    @property
    def yaw(self):
        return self._yaw_pitch_roll[0]

    @yaw.setter
    def yaw(self, yaw: float):
        self._yaw_pitch_roll[0] = yaw
        self._update_orientation()

    @property
    def pitch(self):
        return self._yaw_pitch_roll[1]

    @pitch.setter
    def pitch(self, pitch: float):
        self._yaw_pitch_roll[1] = pitch
        self._update_orientation()

    @property
    def roll(self):
        return self._yaw_pitch_roll[2]

    @roll.setter
    def roll(self, roll: float):
        self._yaw_pitch_roll[2] = roll
        self._update_orientation()

    @property
    def paths(self):
        if not self.parents:
            return [[self]]

        all_paths = []
        for parent in self.parents:
            parent_paths = parent.paths
            for path in parent_paths:
                path.append(self)
            all_paths.extend(parent_paths)

        return all_paths

    @property
    def paths_str(self):
        if not self.parents:
            return ["/" + self.name]

        all_paths_str = []
        for parent in self.parents:
            parent_paths_str = parent.paths_str
            for i, path_str in enumerate(parent_paths_str):
                parent_paths_str[i] = path_str + "/" + self.name
            all_paths_str.extend(parent_paths_str)

        return all_paths_str

    def _clear_dirty_scenes(self, scenes):
        for scene in scenes:
            if scene in self._transform_dirty:
                self._transform_dirty.remove(scene)

            if scene in self._children_transform_dirty:
                del self._children_transform_dirty[scene]

        for child in self._children:
            child._clear_dirty_scenes(scenes)

    def _update_scenes(self):
        old_scenes = self._scenes
        self._scenes = WeakSet()
        for parent in self.parents:
            self._scenes.update(parent._scenes)

        self._clear_dirty_scenes(old_scenes - self._scenes)

    def _set_upstream_dirty(self, scenes):
        for parent in self.parents:
            for scene in scenes:
                if scene not in parent._children_transform_dirty:
                    parent._children_transform_dirty[scene] = set()

                parent._children_transform_dirty[scene].add(self)

            parent._set_upstream_dirty(scenes)

    def _set_transform_dirty(self, scenes):
        if self.__class__.__name__ == "SceneNode":
            return False
        else:
            self._transform_dirty.update(scenes)
            return True

    def _set_dirty(self, self_transform_dirty=True, upstream=True, scenes=None):
        if scenes is None:
            scenes = self._scenes

        self_transform_dirty_set = False
        if self_transform_dirty:
            self_transform_dirty_set = self._set_transform_dirty(scenes)

        for scene in scenes:
            if self._children:
                if scene not in self._children_transform_dirty:
                    self._children_transform_dirty[scene] = set()

                for child in self._children:
                    self._children_transform_dirty[scene].add(child)
                    child._set_dirty(True, False, scenes)
            elif scene in self._children_transform_dirty:
                del self._children_transform_dirty[scene]

        if upstream and (self_transform_dirty_set or self._children):
            self._set_upstream_dirty(scenes)

    @property
    def parent(self):
        for parent in self._parents:
            return parent

    @property
    def scene(self):
        for scene in self._scenes:
            return scene

    @property
    def path(self) -> list:
        if self.parent is None:
            return [self]

        path = self.parent.paths[0]
        path.append(self)
        return path

    def _add_as_child_callback(self) -> None:
        if self._unique_path:
            len_parents = len(self._parents)
            i = 0
            for parent in self._parents:
                if i >= len_parents - 1:
                    break

                parent.remove_child(self)
                i += 1

            node = self
            while True:
                if len(node._parents) > 1:
                    raise RuntimeError("SceneNode can only have one path to root")

                try:
                    node = node._parents[0]
                except IndexError:
                    break

        for child in self._children:
            child._add_as_child_callback()



from .AnimationGroup import AnimationGroup


class SequentialAnimation(AnimationGroup):

    def __init__(self, *animations, **kwargs):
        AnimationGroup.__init__(self, *animations, **kwargs)

    def _update_duration(self):
        if not self._total_duration_dirty:
            return

        self._duration = 0
        for animation in self:
            self._duration += animation.total_duration

        self._total_duration = self._duration * self.loops
        if self._go_back:
            self._total_duration *= 2

        self._total_duration_dirty = False

    def _go_to(self, t: float):
        if t > self.total_duration:
            t = self.total_duration
            self._running = False

        progress = t / self.duration
        int_progress = int(progress)
        if progress != int_progress:
            progress -= int_progress
        elif progress != 0:
            progress = 1
        else:
            progress = 0

        if self.go_back and int_progress % 2 == 1:
            progress = 1 - progress

        reduce_t = progress * self.duration

        accum_time = 0
        next_accum_time = 0
        active_animation = None
        for animation in self:
            if animation is None:
                continue

            next_accum_time += animation.total_duration
            if accum_time <= reduce_t < next_accum_time:
                active_animation = animation
                break
            accum_time = next_accum_time

        if active_animation is None:
            self._running = False
            return

        active_animation._go_to(reduce_t - accum_time)

        if self.running_callback is not None:
            self.running_callback(t)



from enum import Enum
import pytweening


class EasingCurve(Enum):

    Linear = 0
    InQuad = 1
    OutQuad = 2
    InOutQuad = 3
    OutInQuad = 4
    InCubic = 5
    OutCubic = 6
    InOutCubic = 7
    OutInCubic = 8
    InQuart = 9
    OutQuart = 10
    InOutQuart = 11
    OutInQuart = 12
    InQuint = 13
    OutQuint = 14
    InOutQuint = 15
    OutInQuint = 16
    InSine = 17
    OutSine = 18
    InOutSine = 19
    OutInSine = 20
    InExpo = 21
    OutExpo = 22
    InOutExpo = 23
    OutInExpo = 24
    InCirc = 25
    OutCirc = 26
    InOutCirc = 27
    OutInCirc = 28
    InElastic = 29
    OutElastic = 30
    InOutElastic = 31
    OutInElastic = 32
    InBack = 33
    OutBack = 34
    InOutBack = 35
    OutInBack = 36
    InBounce = 37
    OutBounce = 38
    InOutBounce = 39
    OutInBounce = 40

    def __call__(self, progress: float) -> float:
        if self == EasingCurve.Linear:
            return progress

        func_name = "ease" + str(self)[len("EasingCurve.") :]
        func = getattr(pytweening, func_name)
        return func(progress)



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math


def init_Octahedron(cls):
    cls.edge_length = math.sqrt(2)
    cls.base_positions = [
        glm.vec3(-1, 0, 0),  # 0
        glm.vec3(1, 0, 0),  # 1
        glm.vec3(0, -1, 0),  # 2
        glm.vec3(0, 1, 0),  # 3
        glm.vec3(0, 0, -1),  # 4
        glm.vec3(0, 0, 1),  # 5
    ]

    cls.base_indices = [
        glm.uvec3(5, 0, 2),
        glm.uvec3(5, 2, 1),
        glm.uvec3(5, 1, 3),
        glm.uvec3(5, 3, 0),
        glm.uvec3(4, 2, 0),
        glm.uvec3(4, 1, 2),
        glm.uvec3(4, 3, 1),
        glm.uvec3(4, 0, 3),
    ]

    cls.fixed_vertices = []
    for index in cls.base_indices:
        pos0 = cls.base_positions[index[0]]
        pos1 = cls.base_positions[index[1]]
        pos2 = cls.base_positions[index[2]]

        v1 = pos1 - pos0
        v2 = pos2 - pos0
        normal = glm.normalize(glm.cross(v1, v2))

        vertex0 = Vertex(position=pos0, normal=normal, tex_coord=glm.vec3(0.5, 1, 0))
        vertex1 = Vertex(
            position=pos1,
            normal=normal,
            tex_coord=glm.vec3(0.5 - math.sqrt(3) / 4, 0.25, 0),
        )
        vertex2 = Vertex(
            position=pos2,
            normal=normal,
            tex_coord=glm.vec3(0.5 + math.sqrt(3) / 4, 0.25, 0),
        )
        cls.fixed_vertices.append(vertex0)
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)

    return cls


@init_Octahedron
class Octahedron(Mesh):

    @checktype
    def __init__(
        self,
        radius=1,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        back_color: (glm.vec3, glm.vec4) = None,
        stable=False,
        normalize_tex_coord=False,
        name: str = "",
    ):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__stable = stable
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        stable = self.__stable
        normalize_tex_coord = self.__normalize_tex_coord

        quat = glm.quat(1, 0, 0, 0)
        if stable:
            quat_z = glm.quat(math.cos(math.pi / 8), 0, 0, math.sin(math.pi / 8))

            theta = math.acos(math.sqrt(3) / 3)
            quat_x = glm.quat(math.cos(theta / 2), math.sin(theta / 2), 0, 0)

            quat = quat_x * quat_z

        for i, fix_vert in enumerate(Octahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = 2 * radius * (
                    tex_coord - glm.vec3(0.5, 0.5, 0)
                ) * Octahedron.edge_length / math.sqrt(3) + glm.vec3(0.5, 0.5, 0)

            vertices[i] = Vertex(
                position=radius * (quat * fix_vert.position),
                normal=quat * fix_vert.normal,
                tex_coord=tex_coord,
            )

        for i in range(0, len(Octahedron.fixed_vertices), 3):
            indices[int(i / 3)] = glm.uvec3(i, i + 1, i + 2)
            self.generate_temp_TBN(vertices[i], vertices[i + 1], vertices[i + 2])

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    @Mesh.param_setter
    def radius(self, radius: float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord

    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag: bool):
        self.__normalize_tex_coord = flag

    @property
    def is_stable(self):
        return self.__stable

    @is_stable.setter
    @Mesh.param_setter
    def is_stable(self, flag: bool):
        self.__stable = flag



from .ShaderEffect import ShaderEffect
from .. import lut
from glass import sampler2D
from glass.utils import extname

import numpy as np
import os
from OpenGL import GL


class LUTEffect(ShaderEffect):

    def __init__(self, LUT: (str, np.ndarray, sampler2D), contribute: float = 1.0):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(
            self, self_folder + "/../glsl/PostProcessEffects/lut.glsl"
        )

        if isinstance(LUT, str) and extname(LUT) == "cube":
            LUT = lut.cube_to_LUT(LUT)

        if not isinstance(LUT, sampler2D):
            LUT = sampler2D(LUT)

        LUT.wrap = GL.GL_CLAMP_TO_EDGE
        self["LUT"] = LUT
        self["contribute"] = contribute
        self.__LUT = LUT
        self.__contribute = contribute

    @property
    def LUT(self):
        return self.__LUT

    @LUT.setter
    def LUT(self, LUT: (str, np.ndarray, sampler2D)):
        if isinstance(LUT, str) and extname(LUT) == "cube":
            LUT = lut.cube_to_LUT(LUT)

        if not isinstance(LUT, sampler2D):
            LUT = sampler2D(LUT)

        LUT.wrap = GL.GL_CLAMP_TO_EDGE
        self.__LUT = LUT
        self["LUT"] = LUT

    @property
    def contribute(self):
        return self.__contribute

    @contribute.setter
    def contribute(self, contribute: float):
        self.__contribute = contribute
        self["contribute"] = contribute



from enum import Enum
import time


class Chronoscope:

    class Status(Enum):
        Stop = 0
        Running = 1
        Paused = 2

    def __init__(self):
        self._segment_start_time = 0
        self._segment_speed = 1
        self._end_time = 0

        self._status: Chronoscope.Status = Chronoscope.Status.Stop
        self._speed: float = 1

    def start(self):
        if self._status == Chronoscope.Status.Running:
            return

        self._segment_start_time = time.perf_counter()
        self._segment_speed = self._speed
        self._status = Chronoscope.Status.Running

    def stop(self):
        if self._status == Chronoscope.Status.Stop:
            return

        self._end_time = 0
        self._status = Chronoscope.Status.Stop

    def pause(self):
        if self._status != Chronoscope.Status.Running:
            return

        duration = self._segment_speed * (
            time.perf_counter() - self._segment_start_time
        )
        self._end_time += duration
        self._status = Chronoscope.Status.Paused

    def go_to(self, t: float):
        self._segment_start_time = time.perf_counter()
        self._segment_speed = self._speed
        self._end_time = t

        if self._status == Chronoscope.Status.Stop:
            self._status = Chronoscope.Status.Paused

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        if self._status == status:
            return

        if self._status == Chronoscope.Status.Stop:
            if status == Chronoscope.Status.Paused:
                self._status = Chronoscope.Status.Paused
            else:
                self.start()
        elif self._status == Chronoscope.Status.Paused:
            if status == Chronoscope.Status.Stop:
                self.stop()
            else:
                self.start()
        elif self._status == Chronoscope.Status.Running:
            if status == Chronoscope.Status.Paused:
                self.pause()
            else:
                self.stop()

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, speed: float):
        if self._speed == speed:
            return

        self._speed = speed

        if self._status != Chronoscope.Status.Running:
            return

        duration = self._segment_speed * (
            time.perf_counter() - self._segment_start_time
        )
        self._end_time += duration

        self._segment_start_time = time.perf_counter()
        self._segment_speed = self._speed

    def time(self) -> float:
        if self._status != Chronoscope.Status.Running:
            return self._end_time

        duration = self._segment_speed * (
            time.perf_counter() - self._segment_start_time
        )
        return self._end_time + duration



from .PostProcessEffect import PostProcessEffect
from glass import ShaderProgram, FBO, sampler2D, GLInfo, samplerCube, sampler2DArray
from ..Frame import Frame

import os


class FXAAEffect(PostProcessEffect):

    __array_programs = {}
    __program = None
    __cube_program = None

    def __init__(self, internal_format: GLInfo.internal_formats = None):
        PostProcessEffect.__init__(self)

        self._internal_format = internal_format
        self._fbo = None
        self._array_fbo = None
        self._cube_fbo = None

    @staticmethod
    def program():
        if FXAAEffect.__program is None:
            FXAAEffect.__program = ShaderProgram()
            FXAAEffect.__program.compile(Frame.draw_frame_vs)
            FXAAEffect.__program.compile(
                os.path.dirname(os.path.abspath(__file__))
                + "/../glsl/PostProcessEffects/FXAA.fs"
            )
        return FXAAEffect.__program

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D, self._internal_format)
        return self._fbo

    @property
    def array_fbo(self):
        if self._array_fbo is None:
            self._array_fbo = FBO()
            self._array_fbo.attach(0, sampler2DArray, self._internal_format)
        return self._array_fbo

    @property
    def cube_fbo(self):
        if self._cube_fbo is None:
            self._cube_fbo = FBO()
            self._cube_fbo.attach(0, samplerCube, self._internal_format)
        return self._cube_fbo

    def need_pos_info(self) -> bool:
        return False

    def apply(
        self, screen_image: (sampler2D, samplerCube, sampler2DArray)
    ) -> (sampler2D, samplerCube, sampler2DArray):
        if isinstance(screen_image, sampler2D):
            self.fbo.resize(screen_image.width, screen_image.height)
            with self.fbo:
                FXAAEffect.program()["screen_image"] = screen_image
                FXAAEffect.program().draw_triangles(start_index=0, total=6)

            return self.fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.cube_fbo.resize(screen_image.width, screen_image.height)
            with self.cube_fbo:
                FXAAEffect.cube_program()["screen_image"] = screen_image
                FXAAEffect.cube_program().draw_triangles(start_index=0, total=6)

            return self.cube_fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.array_fbo.resize(
                screen_image.width, screen_image.height, layers=screen_image.layers
            )
            with self.array_fbo:
                program = FXAAEffect.array_program(screen_image.layers)
                program["screen_image"] = screen_image
                program.draw_triangles(start_index=0, total=6)

            return self.array_fbo.color_attachment(0)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        FXAAEffect.program()["screen_image"] = screen_image
        FXAAEffect.program().draw_triangles(start_index=0, total=6)



from .CommonRenderer import CommonRenderer
from glass import GLConfig, sampler2D


class ForwardRenderer(CommonRenderer):

    def __init__(self):
        CommonRenderer.__init__(self)

    def startup(self):
        CommonRenderer.startup(self)
        screen = self.camera.screen
        if not screen._samples_set_by_user and not screen._is_gl_init:
            screen._set_samples(4)

    def _draw_opaque(self):
        GLConfig.clear_buffers()

        if self._opaque_meshes:
            self.prepare_forward_draw_mesh(True)
            for mesh, instances in self._opaque_meshes:
                self.forward_draw_mesh(mesh, instances)

        if self._opaque_lines:
            self.prepare_forward_draw_lines(True)
            for mesh, instances in self._opaque_lines:
                self.forward_draw_lines(mesh, instances)

        if self._opaque_points:
            self.prepare_forward_draw_points(True)
            for mesh, instances in self._opaque_points:
                self.forward_draw_points(mesh, instances)

        # 绘制天空盒
        if self.scene.skybox.is_completed:
            self.scene.skybox.draw(self.camera)

        # 绘制天空穹顶
        elif self.scene.skydome.is_completed:
            self.scene.skydome.draw(self.camera)

    def draw_opaque(self):
        if (
            not self._opaque_meshes
            and not self._opaque_lines
            and not self._opaque_points
            and not self.scene.skybox.is_completed
            and not self.scene.skydome.is_completed
        ):
            GLConfig.clear_buffers()
            return

        need_fbo = (
            self._transparent_meshes
            or self._transparent_points
            or self._transparent_lines
            or self.screen._post_process_effects.has_valid
        )

        with GLConfig.LocalConfig(depth_test=True, blend=False):
            if need_fbo:
                with self.OIT_fbo:
                    self._draw_opaque()
            else:
                self._draw_opaque()

        if need_fbo:
            self.OIT_fbo.draw_to_active(0)

            resolved = self.OIT_fbo.resolved
            self._depth_map = resolved.depth_attachment
            self._view_pos_map = resolved.color_attachment(3)
            self._view_normal_map = resolved.color_attachment(4)

    def render(self):
        # profiler.enable()
        self._should_update = False
        sampler2D._should_update = False
        self.classify_meshes()
        self.update_dir_lights_depth()
        self.update_point_lights_depth()
        self.update_spot_lights_depth()
        self.draw_opaque()
        self.draw_transparent()
        # profiler.disable()
        return self._should_update or sampler2D._should_update



from .SkyBox import SkyBox
from .SkyDome import SkyDome

from glass import samplerCube, sampler2D

import glm


class Background:
    def __init__(self):
        self._skybox: SkyBox = SkyBox()
        self._skydome: SkyDome = SkyDome()
        self._color: glm.vec4 = glm.vec4(0, 0, 0, 1)
        self._distance: float = 100.0

    @property
    def skybox(self):
        return self._skybox

    @skybox.setter
    def skybox(self, skybox_map: samplerCube):
        self._skybox.skybox_map = skybox_map

    @property
    def skydome(self):
        return self._skydome

    @skydome.setter
    def skydome(self, image: str):
        self._skydome.skydome_map = image

    @property
    def skybox_map(self) -> samplerCube:
        return self._skybox.skybox_map

    @property
    def skydome_map(self) -> sampler2D:
        return self._skydome.skydome_map

    @property
    def distance(self) -> float:
        return self._distance

    @distance.setter
    def distance(self, distance: float) -> None:
        self._distance = distance

    @property
    def color(self) -> glm.vec4:
        return self._color

    @color.setter
    def color(self, color: (glm.vec4, glm.vec3)) -> None:
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        self._color = color



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame
from glass.utils import checktype
from glass import (
    ShaderProgram,
    FBO,
    sampler2D,
    sampler2DArray,
    samplerCube,
    GLConfig,
    Block,
)

import numpy as np
from OpenGL import GL
import os


class KernelFilter(PostProcessEffect):

    class Kernel(Block.HostClass):
        def __init__(self, kernel: np.ndarray):
            Block.HostClass.__init__(self)

            self._rows = kernel.shape[0]
            self._cols = kernel.shape[1]
            self._data = list(kernel.flatten())
            self._kernel = kernel

        @property
        def rows(self):
            return self._rows

        @property
        def cols(self):
            return self._cols

        @property
        def data(self):
            return self._data

        @property
        def kernel(self):
            return self._kernel

        @kernel.setter
        @Block.HostClass.not_const
        def kernel(self, kernel: np.ndarray):
            self.rows = kernel.shape[0]
            self.cols = kernel.shape[1]
            self.data = list(kernel.flatten())
            self._kernel = kernel

    @checktype
    def __init__(self, kernel: np.ndarray):
        PostProcessEffect.__init__(self)
        self._kernel = KernelFilter.Kernel(kernel)

        self._program = None
        self._cube_program = None
        self._fbo = None
        self._array_fbo = None
        self._cube_fbo = None
        self.__array_programs = {}

    def need_pos_info(self) -> bool:
        return False

    @property
    def program(self):
        if self._program is None:
            self._program = ShaderProgram()
            self._program.compile(Frame.draw_frame_vs)
            self._program.compile(
                os.path.dirname(os.path.abspath(__file__))
                + "/../glsl/PostProcessEffects/kernel_filter.fs"
            )
            self._program["Kernel"].bind(self._kernel)
        return self._program

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D)
        return self._fbo

    @property
    def array_fbo(self):
        if self._array_fbo is None:
            self._array_fbo = FBO()
            self._array_fbo.attach(0, sampler2DArray)
        return self._array_fbo

    @property
    def cube_fbo(self):
        if self._cube_fbo is None:
            self._cube_fbo = FBO()
            self._cube_fbo.attach(0, samplerCube)
        return self._cube_fbo

    def __call__(
        self, screen_image: (sampler2D, sampler2DArray, samplerCube)
    ) -> (sampler2D, sampler2DArray, samplerCube):
        if isinstance(screen_image, sampler2D):
            self.fbo.resize(screen_image.width, screen_image.height)
            with GLConfig.LocalConfig(
                depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL
            ):
                with self.fbo:
                    self.program["screen_image"] = screen_image
                    self.program.draw_triangles(start_index=0, total=6)

            return self.fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.array_fbo.resize(
                screen_image.width, screen_image.height, layers=screen_image.layers
            )
            program = self.array_program(screen_image.layers)
            with GLConfig.LocalConfig(
                depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL
            ):
                with self.array_fbo:
                    program["screen_image"] = screen_image
                    program.draw_triangles(start_index=0, total=6)

            return self.array_fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.cube_fbo.resize(screen_image.width, screen_image.height)
            with GLConfig.LocalConfig(
                depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL
            ):
                with self.cube_fbo:
                    self.cube_program["screen_image"] = screen_image
                    self.cube_program.draw_triangles(start_index=0, total=6)

            return self.cube_fbo.color_attachment(0)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        with GLConfig.LocalConfig(
            depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL
        ):
            self.program["screen_image"] = screen_image
            self.program.draw_triangles(start_index=0, total=6)

    @property
    def kernel(self) -> np.ndarray:
        return self._kernel.kernel

    @kernel.setter
    def kernel(self, kernel: np.ndarray):
        self._kernel.kernel = kernel



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm


class RectFace(Mesh):

    @checktype
    def __init__(
        self,
        width: float = 2,
        height: float = 1,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        back_color: (glm.vec3, glm.vec4) = None,
        vertical: bool = False,
        normalize_tex_coord: bool = False,
        name: str = "",
    ):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__width = width
        self.__height = height
        self.__vertical = vertical
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        width = self.__width
        height = self.__height
        vertical = self.__vertical
        normalize_tex_coord = self.__normalize_tex_coord

        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        # 左下
        vertex0 = Vertex()
        vertex0.position = glm.vec3(-width / 2, -height / 2, 0)
        if vertical:
            vertex0.position = glm.vec3(-width / 2, 0, -height / 2)
        vertex0.normal = normal
        vertex0.tex_coord = glm.vec3(0, 1 - height, 0)
        if normalize_tex_coord:
            vertex0.tex_coord = glm.vec3(0, 1 - height / width, 0)

        # 右下
        vertex1 = Vertex()
        vertex1.position = glm.vec3(width / 2, -height / 2, 0)
        if vertical:
            vertex1.position = glm.vec3(width / 2, 0, -height / 2)
        vertex1.normal = normal
        vertex1.tex_coord = glm.vec3(width, 1 - height, 0)
        if normalize_tex_coord:
            vertex1.tex_coord = glm.vec3(1, 1 - height / width, 0)

        # 右上
        vertex2 = Vertex()
        vertex2.position = glm.vec3(width / 2, height / 2, 0)
        if vertical:
            vertex2.position = glm.vec3(width / 2, 0, height / 2)
        vertex2.normal = normal
        vertex2.tex_coord = glm.vec3(width, 1, 0)
        if normalize_tex_coord:
            vertex2.tex_coord = glm.vec3(1, 1, 0)

        # 左上
        vertex3 = Vertex()
        vertex3.position = glm.vec3(-width / 2, height / 2, 0)
        if vertical:
            vertex3.position = glm.vec3(-width / 2, 0, height / 2)
        vertex3.normal = normal
        vertex3.tex_coord = glm.vec3(0, 1, 0)

        vertices[0] = vertex0
        vertices[1] = vertex1
        vertices[2] = vertex2
        vertices[3] = vertex3

        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        self.generate_temp_TBN(vertices[0], vertices[1], vertices[2])
        self.generate_temp_TBN(vertices[0], vertices[2], vertices[3])

        del vertices[4:]
        del indices[2:]

    @property
    def vertical(self):
        return self.__vertical

    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag: bool):
        self.__vertical = flag

    @property
    def width(self):
        return self.__width

    @width.setter
    @Mesh.param_setter
    def width(self, width: float):
        self.__width = width

    @property
    def height(self):
        return self.__height

    @height.setter
    @Mesh.param_setter
    def height(self, height: float):
        self.__height = height

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord

    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag: bool):
        self.__normalize_tex_coord = flag



from .Icosahedron import Icosahedron

from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math
import copy


class Icosphere(Mesh):

    __base_vertices = []
    __base_indices = []

    @checktype
    def __init__(
        self,
        radius: float = 1,
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        back_color: (glm.vec3, glm.vec4) = None,
        n_levels: int = 4,
        name: str = "",
        block=True,
    ):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__radius = radius
        self.__n_levels = n_levels
        self.start_building()

    def __generate_base_data(self):
        if Icosphere.__base_indices:
            return

        Icosphere.__base_indices = copy.deepcopy(Icosahedron.base_indices)
        Icosphere.__base_vertices = []

        half_theta = math.atan(2) / 4
        cos_half_theta = math.cos(half_theta)
        sin_half_theta = math.sin(half_theta)
        q = glm.quat(cos_half_theta, 0, sin_half_theta, 0)
        for base_pos in Icosahedron.base_positions:
            pos = q * base_pos
            if glm.length(pos - glm.vec3(0, 0, 1)) < 1e-6:
                pos = glm.vec3(0, 0, 1)
            elif glm.length(pos - glm.vec3(0, 0, -1)) < 1e-6:
                pos = glm.vec3(0, 0, -1)
            Icosphere.__base_vertices.append(Icosphere.__create_vertex(pos))

        for index in Icosphere.__base_indices:
            vertex0 = Icosphere.__base_vertices[index[0]]
            vertex1 = Icosphere.__base_vertices[index[1]]
            vertex2 = Icosphere.__base_vertices[index[2]]

            v1 = vertex1.position - vertex0.position
            v2 = vertex2.position - vertex0.position
            normal = glm.cross(v1, v2)

            for i in range(3):
                vertex = Icosphere.__base_vertices[index[i]]
                if vertex.tex_coord.t == 0 or vertex.tex_coord.t == 1:
                    new_vertex = Icosphere.__create_vertex(vertex.position, normal)
                    new_index = len(Icosphere.__base_vertices)
                    Icosphere.__base_vertices.append(new_vertex)
                    index[i] = new_index

    def build(self):
        self.is_closed = True
        vertices = self.vertices
        indices = self.indices
        self.__generate_base_data()

        radius = self.__radius
        n_levels = self.__n_levels

        _indices = Icosphere.__base_indices

        i_vertex = 0
        i_index = 0
        for vertex in Icosphere.__base_vertices:
            new_vertex = copy.deepcopy(vertex)
            new_vertex.position = radius * new_vertex.position
            new_vertex.tangent = radius * new_vertex.tangent
            new_vertex.bitangent = radius * new_vertex.bitangent
            vertices[i_vertex] = new_vertex
            i_vertex += 1

        for _ in range(n_levels):
            i_index = 0
            for index in _indices:
                vertex0 = vertices[index[0]]
                vertex1 = vertices[index[1]]
                vertex2 = vertices[index[2]]

                new_pos0 = glm.normalize(vertex1.position + vertex2.position)
                new_pos1 = glm.normalize(vertex0.position + vertex2.position)
                new_pos2 = glm.normalize(vertex0.position + vertex1.position)

                new_vertex0 = Icosphere.__create_vertex(new_pos0, radius=radius)
                new_vertex1 = Icosphere.__create_vertex(new_pos1, radius=radius)
                new_vertex2 = Icosphere.__create_vertex(new_pos2, radius=radius)

                start_index = i_vertex
                vertices[i_vertex] = new_vertex0
                i_vertex += 1

                vertices[i_vertex] = new_vertex1
                i_vertex += 1

                vertices[i_vertex] = new_vertex2
                i_vertex += 1

                indices[i_index] = glm.uvec3(index[0], start_index + 2, start_index + 1)
                i_index += 1

                indices[i_index] = glm.uvec3(index[1], start_index, start_index + 2)
                i_index += 1

                indices[i_index] = glm.uvec3(index[2], start_index + 1, start_index)
                i_index += 1

                indices[i_index] = glm.uvec3(
                    start_index, start_index + 1, start_index + 2
                )
                i_index += 1

                yield

            _indices = list(indices)

        del vertices[i_vertex:]
        del indices[i_index:]

        for index in indices:
            vertex0 = vertices[index[0]]
            vertex1 = vertices[index[1]]
            vertex2 = vertices[index[2]]
            s0 = vertex0.tex_coord.s
            s1 = vertex1.tex_coord.s
            s2 = vertex2.tex_coord.s

            if abs(s0 - s1) > 0.5 and abs(s0 - s2) > 0.5:
                new_vertex = copy.deepcopy(vertex0)
                new_vertex.tex_coord.s += 1 if s1 + s2 > 1 else -1
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[0] = new_index

            if abs(s1 - s0) > 0.5 and abs(s1 - s2) > 0.5:
                new_vertex = copy.deepcopy(vertex1)
                new_vertex.tex_coord.s += 1 if s0 + s2 > 1 else -1
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[1] = new_index

            if abs(s2 - s0) > 0.5 and abs(s2 - s1) > 0.5:
                new_vertex = copy.deepcopy(vertex2)
                new_vertex.tex_coord.s += 1 if s0 + s1 > 1 else -1
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[2] = new_index

            yield

    def __get_info(position, normal=None):
        dem = math.sqrt(position.x * position.x + position.y * position.y)
        if dem == 0:
            s = 0
            tangent = glm.vec3(0, 2 * math.pi, 0)
            bitangent = glm.vec3(-math.pi, 0, 0)
            if normal is not None:
                theta = math.atan2(normal.y, normal.x)
                phi = math.pi / 2 if normal.z > 0 else -math.pi / 2
                sin_theta = math.sin(theta)
                cos_theta = math.cos(theta)
                sin_phi = math.sin(phi)
                cos_phi = math.cos(phi)
                s = theta / (2 * math.pi) + 0.5
                tangent = 2 * math.pi * cos_phi * glm.vec3(-sin_theta, cos_theta, 0)
                bitangent = math.pi * glm.vec3(
                    -sin_phi * cos_theta, -sin_phi * sin_theta, cos_phi
                )
            t = int((position.z + 1) / 2)

            return glm.vec3(s, t, 0), tangent, bitangent

        theta = math.atan2(position.y, position.x)
        phi = math.atan(position.z / dem)
        sin_theta = math.sin(theta)
        cos_theta = math.cos(theta)
        sin_phi = math.sin(phi)
        cos_phi = math.cos(phi)

        s = theta / (2 * math.pi) + 0.5
        t = phi / math.pi + 0.5

        tex_coord = glm.vec3(s, t, 0)
        tangent = 2 * math.pi * cos_phi * glm.vec3(-sin_theta, cos_theta, 0)
        bitangent = math.pi * glm.vec3(
            -sin_phi * cos_theta, -sin_phi * sin_theta, cos_phi
        )
        return tex_coord, tangent, bitangent

    @staticmethod
    def __create_vertex(position, normal=None, radius=1, **kwargs):
        tex_coord, tangent, bitangent = Icosphere.__get_info(position, normal)
        return Vertex(
            position=radius * position,
            tangent=radius * tangent,
            bitangent=radius * bitangent,
            normal=position,
            tex_coord=tex_coord,
            **kwargs
        )

    @property
    def n_levels(self):
        return self.__n_levels

    @n_levels.setter
    @Mesh.param_setter
    def n_levels(self, n_levels: int):
        self.__n_levels = n_levels

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    @Mesh.param_setter
    def radius(self, radius: float):
        self.__radius = radius



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass import (
    FBO,
    ShaderProgram,
    sampler2D,
    GLConfig,
    sampler2DArray,
    samplerCube,
    GLInfo,
)
from glass.utils import checktype
from glass.helper import get_channels

from OpenGL import GL
import glm
import os


class GaussBlur(PostProcessEffect):

    __array_programs = {}
    __program = None
    __cube_program = None

    @staticmethod
    def program():
        if GaussBlur.__program is None:
            GaussBlur.__program = ShaderProgram()
            GaussBlur.__program.compile(Frame.draw_frame_vs)
            GaussBlur.__program.compile(
                os.path.dirname(os.path.abspath(__file__))
                + "/../glsl/PostProcessEffects/gauss_blur.fs"
            )
        return GaussBlur.__program

    @checktype
    def __init__(
        self,
        kernel_shape: (int, tuple) = 32,
        sigma: (float, tuple) = 0,
        internal_format: GLInfo.internal_formats = GL.GL_RGBA32F,
    ):
        PostProcessEffect.__init__(self)

        self.__kernel_shape = glm.uvec2()
        self.__sigma = glm.vec2()
        self.__channels = get_channels(internal_format)
        self.__internal_format = internal_format

        if isinstance(kernel_shape, tuple):
            self.__kernel_shape.x, self.__kernel_shape.y = kernel_shape
        else:
            self.__kernel_shape.x, self.__kernel_shape.y = kernel_shape, kernel_shape

        if isinstance(sigma, tuple):
            self.__sigma.x, self.__sigma.y = sigma
        else:
            self.__sigma.x, self.__sigma.y = sigma, sigma

        self._horizontal_fbo = None
        self._vertical_fbo = None
        self._horizontal_array_fbo = None
        self._vertical_array_fbo = None
        self._horizontal_cube_fbo = None
        self._vertical_cube_fbo = None

    def need_pos_info(self):
        return False

    @property
    def horizontal_fbo(self):
        if self._horizontal_fbo is None:
            self._horizontal_fbo = FBO()
            self._horizontal_fbo.attach(0, sampler2D, self.__internal_format)
        return self._horizontal_fbo

    @property
    def vertical_fbo(self):
        if self._vertical_fbo is None:
            self._vertical_fbo = FBO()
            self._vertical_fbo.attach(0, sampler2D, self.__internal_format)
        return self._vertical_fbo

    @property
    def horizontal_array_fbo(self):
        if self._horizontal_array_fbo is None:
            self._horizontal_array_fbo = FBO()
            self._horizontal_array_fbo.attach(0, sampler2DArray, self.__internal_format)
        return self._horizontal_array_fbo

    @property
    def vertical_array_fbo(self):
        if self._vertical_array_fbo is None:
            self._vertical_array_fbo = FBO()
            self._vertical_array_fbo.attach(0, sampler2DArray, self.__internal_format)
        return self._vertical_array_fbo

    @property
    def horizontal_cube_fbo(self):
        if self._horizontal_cube_fbo is None:
            self._horizontal_cube_fbo = FBO()
            self._horizontal_cube_fbo.attach(0, samplerCube, self.__internal_format)
        return self._horizontal_cube_fbo

    @property
    def vertical_cube_fbo(self):
        if self._vertical_cube_fbo is None:
            self._vertical_cube_fbo = FBO()
            self._vertical_cube_fbo.attach(0, samplerCube, self.__internal_format)
        return self._vertical_cube_fbo

    def apply(
        self, screen_image: (sampler2D, sampler2DArray, samplerCube)
    ) -> (sampler2D, sampler2DArray, samplerCube):
        if self.__sigma.x == 0:
            self.__sigma.x = 0.3 * ((self.__kernel_shape.x - 1) * 0.5 - 1) + 0.8
        if self.__sigma.y == 0:
            self.__sigma.y = 0.3 * ((self.__kernel_shape.y - 1) * 0.5 - 1) + 0.8

        if isinstance(screen_image, sampler2D):
            self.horizontal_fbo.resize(screen_image.width, screen_image.height)
            self.vertical_fbo.resize(screen_image.width, screen_image.height)

            program = GaussBlur.program()
            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(start_index=0, total=6)

                with self.vertical_fbo:
                    program["screen_image"] = self.horizontal_fbo.color_attachment(0)
                    program["horizontal"] = False
                    program.draw_triangles(start_index=0, total=6)

            return self.vertical_fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.horizontal_array_fbo.resize(
                screen_image.width, screen_image.height, layers=screen_image.layers
            )
            self.vertical_array_fbo.resize(
                screen_image.width, screen_image.height, layers=screen_image.layers
            )

            program = GaussBlur.array_program(screen_image.layers)

            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_array_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(start_index=0, total=6)

                with self.vertical_array_fbo:
                    program["screen_image"] = (
                        self.horizontal_array_fbo.color_attachment(0)
                    )
                    program["horizontal"] = False
                    program.draw_triangles(start_index=0, total=6)

            return self.vertical_array_fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.horizontal_cube_fbo.resize(screen_image.width, screen_image.height)
            self.vertical_cube_fbo.resize(screen_image.width, screen_image.height)

            program = GaussBlur.cube_program()

            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_cube_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(start_index=0, total=6)

                with self.vertical_cube_fbo:
                    program["screen_image"] = self.horizontal_cube_fbo.color_attachment(
                        0
                    )
                    program["horizontal"] = False
                    program.draw_triangles(start_index=0, total=6)

            return self.vertical_cube_fbo.color_attachment(0)

    def draw_to_active(self, screen_image: sampler2D) -> None:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        if self.__sigma.x == 0:
            self.__sigma.x = ((self.__kernel_shape.x - 1) * 0.5 - 1) / 3
        if self.__sigma.y == 0:
            self.__sigma.y = ((self.__kernel_shape.y - 1) * 0.5 - 1) / 3

        program = GaussBlur.program()

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

            program["kernel_shape"] = self.__kernel_shape
            program["sigma"] = self.__sigma
            program["channels"] = self.__channels

            with self.horizontal_fbo:
                program["screen_image"] = screen_image
                program["horizontal"] = True
                program.draw_triangles(start_index=0, total=6)

            GLConfig.clear_buffers()
            program["screen_image"] = self.horizontal_fbo.color_attachment(0)
            program["horizontal"] = False
            program.draw_triangles(start_index=0, total=6)

    @property
    def kernel_shape(self) -> tuple:
        return (self.__kernel_shape.x, self.__kernel_shape.y)

    @kernel_shape.setter
    @checktype
    def kernel_shape(self, shape: (int, tuple)):
        if isinstance(shape, tuple):
            self.__kernel_shape.x, self.__kernel_shape.y = shape
        else:
            self.__kernel_shape.x, self.__kernel_shape.y = shape, shape

    @property
    def kernel_width(self) -> int:
        return self.__kernel_shape.x

    @kernel_width.setter
    @checktype
    def kernel_width(self, width: int):
        self.__kernel_shape.x = width

    @property
    def kernel_height(self) -> int:
        return self.__kernel_shape.y

    @kernel_height.setter
    @checktype
    def kernel_height(self, height: int):
        self.__kernel_shape.y = height

    @property
    def sigma(self) -> tuple:
        return (self.__sigma.x, self.__sigma.y)

    @sigma.setter
    @checktype
    def sigma(self, sigma: (float, tuple)):
        if isinstance(sigma, tuple):
            self.__sigma.x, self.__sigma.y = sigma
        else:
            self.__sigma.x, self.__sigma.y = sigma, sigma

    @property
    def sigma_x(self) -> float:
        return self.__sigma.x

    @sigma_x.setter
    def sigma_x(self, value: float):
        self.__sigma.x = value

    @property
    def sigma_y(self) -> float:
        return self.__sigma.y

    @sigma_y.setter
    def sigma_y(self, value: float):
        self.__sigma.y = value



from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
from OpenGL import GL


class Polyline(Mesh):

    def __init__(
        self,
        points: list = [],
        color: (glm.vec3, glm.vec4) = glm.vec4(0.396, 0.74151, 0.69102, 1),
        line_width: int = 2,
        loop: bool = False,
        name: str = "",
        block: bool = True,
    ):
        primitive_type = GL.GL_LINE_STRIP if not loop else GL.GL_LINE_LOOP
        Mesh.__init__(
            self,
            primitive_type=primitive_type,
            color=color,
            name=name,
            block=block,
            shared=False,
        )
        self.render_hints.line_width = line_width
        self.__points = points
        self.start_building()

    def build(self):
        points = self.__points
        vertices = self.vertices

        length = 0
        len_points = len(points)
        for i in range(len_points):
            if i > 0:
                length += glm.length(points[i] - points[i - 1])
            bitangent = None
            if i < len_points - 1:
                bitangent = glm.normalize(points[i + 1] - points[i])
            else:
                bitangent = glm.normalize(points[i] - points[i - 1])
            vertices[i] = Vertex(
                position=points[i],
                bitangent=bitangent,
                tex_coord=glm.vec3(length, 0, 0),
            )

        del vertices[len_points:]

    @property
    def points(self):
        return self.__points

    @points.setter
    def points(self, points):
        self.__points = points
        self.start_building()

    @property
    def line_width(self):
        return self.render_hints.line_width

    @line_width.setter
    @checktype
    def line_width(self, line_width: int):
        self.render_hints.line_width = line_width

    @property
    def loop(self):
        return self.primitive_type == GL.GL_LINE_LOOP

    @loop.setter
    @Mesh.param_setter
    def loop(self, flag: bool):
        self.primitive_type = GL.GL_LINE_STRIP if not flag else GL.GL_LINE_LOOP
